""" This is the StorageElement class.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# # custom duty

import six
import copy
import datetime
import errno
import os
import re
import sys
import threading
import time


from functools import reduce

# # from DIRAC
from DIRAC import gLogger, gConfig, siteName
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.File import convertSizeUnits
from DIRAC.Core.Utilities.List import getIndexInList
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR, returnSingleResult
from DIRAC.Resources.Storage.StorageFactory import StorageFactory
from DIRAC.Core.Utilities.Pfn import pfnparse
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.Core.Security.Locations import getProxyLocation
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.AccountingSystem.Client.Types.DataOperation import DataOperation
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

__RCSID__ = "$Id$"

DEFAULT_OCCUPANCY_FILE = "occupancy.json"

sLog = gLogger.getSubLogger(__name__)


class StorageElementCache(object):
    def __init__(self):
        self.seCache = DictCache()

    def __call__(self, name, plugins=None, vo=None, hideExceptions=False):
        self.seCache.purgeExpired(expiredInSeconds=60)
        tId = threading.current_thread().ident

        if not vo:
            result = getVOfromProxyGroup()
            if not result["OK"]:
                return
            vo = result["Value"]

        # Because the gfal2 context caches the proxy location,
        # we also use the proxy location as a key.
        # In practice, there should almost always be one, except for the REA
        # If we see its memory consumtpion exploding, this might be a place to look
        proxyLoc = getProxyLocation()

        # ensure plugins is hashable! (tuple)
        if isinstance(plugins, list):
            plugins = tuple(plugins)

        argTuple = (tId, name, plugins, vo, proxyLoc)
        seObj = self.seCache.get(argTuple)

        if not seObj:
            seObj = StorageElementItem(name, plugins, vo, hideExceptions=hideExceptions)
            # Add the StorageElement to the cache for 1/2 hour
            self.seCache.add(argTuple, 1800, seObj)

        return seObj


class StorageElementItem(object):
    """
    .. class:: StorageElement

    common interface to the grid storage element



      self.name is the resolved name of the StorageElement i.e CERN-tape
      self.options is dictionary containing the general options defined in the CS e.g. self.options['Backend] = 'Castor2'
      self.storages is a list of the stub objects created by StorageFactory for the protocols found in the CS.
      self.localPlugins is a list of the local protocols that were created by StorageFactory
      self.remotePlugins is a list of the remote protocols that were created by StorageFactory
      self.protocolOptions is a list of dictionaries containing the options found in the CS. (should be removed)



    dynamic method::

      retransferOnlineFile( lfn )
      exists( lfn )
      isFile( lfn )
      getFile( lfn, localPath = False )
      putFile( lfnLocal, sourceSize = 0 ) : {lfn:local}
      replicateFile( lfn, sourceSize = 0 )
      getFileMetadata( lfn )
      getFileSize( lfn )
      removeFile( lfn )
      prestageFile( lfn, lifetime = 86400 )
      prestageFileStatus( lfn )
      pinFile( lfn, lifetime = 60 * 60 * 24 )
      releaseFile( lfn )
      isDirectory( lfn )
      getDirectoryMetadata( lfn )
      getDirectorySize( lfn )
      listDirectory( lfn )
      removeDirectory( lfn, recursive = False )
      createDirectory( lfn )
      putDirectory( lfn )
      getDirectory( lfn, localPath = False )


    """

    __deprecatedArguments = ["singleFile", "singleDirectory"]  # Arguments that are now useless

    # Some methods have a different name in the StorageElement and the plugins...
    # We could avoid this static list in the __getattr__ by checking the storage plugin and so on
    # but fine... let's not be too smart, otherwise it becomes unreadable :-)
    __equivalentMethodNames = {
        "exists": "exists",
        "isFile": "isFile",
        "getFile": "getFile",
        "putFile": "putFile",
        "replicateFile": "putFile",
        "getFileMetadata": "getFileMetadata",
        "getFileSize": "getFileSize",
        "removeFile": "removeFile",
        "prestageFile": "prestageFile",
        "prestageFileStatus": "prestageFileStatus",
        "pinFile": "pinFile",
        "releaseFile": "releaseFile",
        "isDirectory": "isDirectory",
        "getDirectoryMetadata": "getDirectoryMetadata",
        "getDirectorySize": "getDirectorySize",
        "listDirectory": "listDirectory",
        "removeDirectory": "removeDirectory",
        "createDirectory": "createDirectory",
        "putDirectory": "putDirectory",
        "getDirectory": "getDirectory",
    }

    # We can set default argument in the __executeFunction which impacts all plugins
    __defaultsArguments = {
        "putFile": {"sourceSize": 0},
        "getFile": {"localPath": False},
        "prestageFile": {"lifetime": 86400},
        "pinFile": {"lifetime": 60 * 60 * 24},
        "removeDirectory": {"recursive": False},
        "getDirectory": {"localPath": False},
    }

    def __init__(self, name, plugins=None, vo=None, hideExceptions=False):
        """c'tor

        :param str name: SE name
        :param list plugins: requested storage plugins
        :param vo: vo

        """

        self.methodName = None

        if plugins is None:
            plugins = []

        if vo:
            self.vo = vo
        else:
            result = getVOfromProxyGroup()
            if not result["OK"]:
                return
            self.vo = result["Value"]
        self.opHelper = Operations(vo=self.vo)

        # These things will soon have to go as well. 'AccessProtocol.1' is all but flexible.
        proxiedProtocols = gConfig.getValue("/LocalSite/StorageElements/ProxyProtocols", "").split(",")
        self.useProxy = (
            gConfig.getValue("/Resources/StorageElements/%s/AccessProtocol.1/Protocol" % name, "UnknownProtocol")
            in proxiedProtocols
        )

        if not self.useProxy:
            self.useProxy = gConfig.getValue("/LocalSite/StorageElements/%s/UseProxy" % name, False)
        if not self.useProxy:
            self.useProxy = self.opHelper.getValue("/Services/StorageElements/%s/UseProxy" % name, False)

        self.valid = True

        res = StorageFactory(useProxy=self.useProxy, vo=self.vo).getStorages(
            name, pluginList=plugins, hideExceptions=hideExceptions
        )

        if not res["OK"]:
            self.valid = False
            self.name = name
            self.errorReason = res["Message"]
        else:
            factoryDict = res["Value"]
            self.name = factoryDict["StorageName"]
            self.options = factoryDict["StorageOptions"]
            self.localPlugins = factoryDict["LocalPlugins"]
            self.remotePlugins = factoryDict["RemotePlugins"]
            self.storages = factoryDict["StorageObjects"]
            self.protocolOptions = factoryDict["ProtocolOptions"]
            self.turlProtocols = factoryDict["TurlProtocols"]

            for storage in self.storages:

                storage.setStorageElement(self)

        self.log = sLog.getSubLogger("SE[%s]" % self.name)

        if self.valid:

            self.useCatalogURL = gConfig.getValue("/Resources/StorageElements/%s/UseCatalogURL" % self.name, False)
            self.log.debug("useCatalogURL: %s" % self.useCatalogURL)

            self.__dmsHelper = DMSHelpers(vo=vo)

            # Allow SE to overwrite general operation config
            accessProto = self.options.get("AccessProtocols")
            self.localAccessProtocolList = accessProto if accessProto else self.__dmsHelper.getAccessProtocols()
            self.log.debug("localAccessProtocolList %s" % self.localAccessProtocolList)

            writeProto = self.options.get("WriteProtocols")
            self.localWriteProtocolList = writeProto if writeProto else self.__dmsHelper.getWriteProtocols()
            self.log.debug("localWriteProtocolList %s" % self.localWriteProtocolList)

            # For the staging protocols, we take in order:
            # * the locally defined staging protocols list
            # * the locally defined access protocols list
            # * the globally defined staging protocols list
            # * the globally defined access protocols list
            # This ensures local over global preference as well
            # as backward compatibility (when staging was part of read methods)

            stageProto = self.options.get("StageProtocols")
            globalStageProto = self.__dmsHelper.getStageProtocols()
            self.localStageProtocolList = (
                stageProto
                if stageProto
                else accessProto
                if accessProto
                else globalStageProto
                if globalStageProto
                else self.localAccessProtocolList
            )
            self.log.debug("localStageProtocolList %s" % self.localStageProtocolList)

        self.stageMethods = ["prestageFile", "prestageFileStatus"]

        self.readMethods = ["getFile", "getDirectory"]

        self.writeMethods = [
            "retransferOnlineFile",
            "putFile",
            "replicateFile",
            "pinFile",
            "releaseFile",
            "createDirectory",
            "putDirectory",
        ]

        self.removeMethods = ["removeFile", "removeDirectory"]

        self.checkMethods = [
            "exists",
            "getDirectoryMetadata",
            "getDirectorySize",
            "getFileSize",
            "getFileMetadata",
            "listDirectory",
            "isDirectory",
            "isFile",
            "getOccupancy",
        ]

        self.okMethods = [
            "getLocalProtocols",
            "getProtocols",
            "getRemoteProtocols",
            "storageElementName",
            "getStorageParameters",
            "getTransportURL",
            "isLocalSE",
        ]

        self.__fileCatalog = None

    def dump(self):
        """Dump to the logger a summary of the StorageElement items."""
        log = self.log.getLocalSubLogger("dump")
        log.debug("Preparing dump for StorageElement %s." % self.name)
        if not self.valid:
            log.debug("Failed to create StorageElement plugins.", self.errorReason)
            return
        i = 1
        outStr = "\n\n============ Options ============\n"
        for key in sorted(self.options):
            outStr = "%s%s: %s\n" % (outStr, key.ljust(15), self.options[key])

        for storage in self.storages:
            outStr = "%s============Protocol %s ============\n" % (outStr, i)
            storageParameters = storage.getParameters()
            for key in sorted(storageParameters):
                outStr = "%s%s: %s\n" % (outStr, key.ljust(15), storageParameters[key])
            i = i + 1
        log.debug(outStr)

    #################################################################################################
    #
    # These are the basic get functions for storage configuration
    #

    def getStorageElementName(self):
        """SE name getter for backward compatibility"""
        return S_OK(self.storageElementName())

    def storageElementName(self):
        """SE name getter"""
        return self.name

    def getChecksumType(self):
        """Checksum type getter for backward compatibility"""
        return S_OK(self.checksumType())

    def checksumType(self):
        """get specific /Resources/StorageElements/<SEName>/ChecksumType option if defined, otherwise
        global /Resources/StorageElements/ChecksumType
        """
        return (
            self.options["ChecksumType"].upper()
            if "ChecksumType" in self.options
            else gConfig.getValue("/Resources/StorageElements/ChecksumType", "ADLER32").upper()
        )

    def getStatus(self):
        """
        Return Status of the SE only if the SE is valid
        It returns an S_OK/S_ERROR structure
        """
        valid = self.isValid()
        if not valid["OK"]:
            return valid
        return S_OK(self.status())

    def isSameSE(self, otherSE):
        """Compares two SE together and tries to guess if the two SEs are pointing at the same
        location from the namespace point of view.
        This is primarily aimed at avoiding to overwrite a file with itself, in particular
        where the difference is only the SRM spacetoken.

        Two SEs are considered to be the same if they have a couple (Host, Path) in common
        among their various protocols

        :param otherSE: the storage element to which we compare
        :returns: boolean. True if the two SEs are the same.
        """

        # If the two objects are the same, it is obviously the same SE
        if self == otherSE:
            return True

        # Otherwise, we build the list of (Host, Path) couples

        selfEndpoints = set()
        otherSEEndpoints = set()

        for storage in self.storages:
            storageParam = storage.getParameters()
            selfEndpoints.add((storageParam["Host"], storageParam["Path"]))

        for storage in otherSE.storages:
            storageParam = storage.getParameters()
            otherSEEndpoints.add((storageParam["Host"], storageParam["Path"]))

        # The two SEs are the same if they have at least one couple in common
        return bool(selfEndpoints & otherSEEndpoints)

    def getOccupancy(self, unit="MB", **kwargs):
        """Retrieves the space information about the storage.
        It returns the Total and Free space, and a SpaceReservation.

        The SpaceReservation is just a name of a zone of the physical storage which can have some space reserved.
        It corresponds to the ``SpaceToken`` concept of SRM.
        If the StorageElement definition has a ``SpaceReservation`` option in the CS, this is returned, unless
        it is overwritten by the storage plugin.

        It loops over the different Storage Plugins to query it.

        :params occupancyLFN: (named param) LFN where to find the space reporting json file on the storage
                              The json file should contain the Free and Total space in B.
                              If not specified, the default path will be </vo/occupancy.json>

        :params unit: (default MB)unit of the value returned. See :py:func:`~DIRAC.Core.Utilities.File.convertSizeUnits`
                      CAUTION: only the `Total` and `Free` field are converted !
                      Since the rest is whatever is returned by the plugin no conversion is performed

        :returns: S_OK with dict (keys: Total, Free, SpaceReservation)
        """
        log = self.log.getSubLogger("getOccupancy")

        if "occupancyLFN" not in kwargs:
            occupancyLFN = self.options.get("OccupancyLFN")
            if not occupancyLFN:
                occupancyLFN = os.path.join("/", self.vo, DEFAULT_OCCUPANCY_FILE)

            kwargs["occupancyLFN"] = occupancyLFN

        filteredPlugins = self.__filterPlugins("getOccupancy")
        if not filteredPlugins:
            return S_ERROR(errno.EPROTONOSUPPORT, "No storage plugins to query the occupancy")

        # Call occupancy plugin if requested
        occupancyPlugin = self.options.get("OccupancyPlugin")
        if occupancyPlugin:
            res = ObjectLoader().loadObject("Resources.Storage.OccupancyPlugins.%s" % occupancyPlugin)
            if not res["OK"]:
                return S_ERROR(errno.EPROTONOSUPPORT, "Failed to load occupancy plugin %s" % occupancyPlugin)
            log.debug("Use occupancy plugin %s" % occupancyPlugin)
            try:
                occupancyPlugin = res["Value"](self)
                res = occupancyPlugin.getOccupancy(**kwargs)
                if not res["OK"]:
                    return res
                occupancyDict = res["Value"]
                return self.checkOccupancy(occupancyDict, unit)
            except Exception as e:
                return S_ERROR("Occupancy plugin failed: %s" % str(e))

        # Try all of the storages one by one
        for storage in filteredPlugins:

            # The result of the plugin is always in B
            res = storage.getOccupancy(**kwargs)
            if res["OK"]:
                occupancyDict = res["Value"]
                result = self.checkOccupancy(occupancyDict, unit)
                if not result["OK"]:
                    continue
                occupancyDict = result["Value"]

                if "SpaceReservation" not in occupancyDict:
                    occupancyDict["SpaceReservation"] = self.options.get("SpaceReservation")
                return S_OK(occupancyDict)

        return S_ERROR("Could not retrieve the occupancy from any plugin")

    def checkOccupancy(self, occupancyDict, unit):
        """Validate occupancy dict given by getOccupancy

        :param dict occupancyDict: occupancy given by occupancy or storage plugins
        :param str unit: Any of ( 'B', 'kB', 'MB', 'GB', 'TB', 'PB')

        :returns: S_OK with updated occupancyDict
        """
        log = self.log.getSubLogger("checkOccupancy")

        # Mandatory parameters
        mandatoryParams = set(["Total", "Free"])
        # Make sure all the mandatory parameters are present
        if set(occupancyDict) & mandatoryParams != mandatoryParams:

            msg = "Missing mandatory parameters %s" % str(mandatoryParams - set(occupancyDict))
            log.error(msg)
            return S_ERROR(msg)

        # Since plugins return Bytes, we do not need to convert if that's what we want
        if unit != "B":
            for space in ["Total", "Free"]:
                convertedSpace = convertSizeUnits(occupancyDict[space], "B", unit)
                # If we have a conversion error, we go to the next plugin
                if convertedSpace == -sys.maxsize:
                    msg = "Error converting %s space from MB to %s: %s" % (space, unit, occupancyDict[space])
                    log.error(msg)
                    return S_ERROR("Error converting")
                occupancyDict[space] = convertedSpace
        return S_OK(occupancyDict)

    def status(self):
        """
         Return Status of the SE, a dictionary with:

          * Read: True (is allowed), False (it is not allowed)
          * Write: True (is allowed), False (it is not allowed)
          * Remove: True (is allowed), False (it is not allowed)
          * Check: True (is allowed), False (it is not allowed).

            .. note:: Check is always allowed IF Read is allowed
                      (regardless of what set in the Check option of the configuration)

          * DiskSE: True if TXDY with Y > 0 (defaults to True)
          * TapeSE: True if TXDY with X > 0 (defaults to False)
          * TotalCapacityTB: float (-1 if not defined)
          * DiskCacheTB: float (-1 if not defined)

        It returns directly the dictionary
        """

        self.log.getSubLogger("getStatus").debug("determining status of %s." % self.name)

        retDict = {}
        if not self.valid:
            retDict["Read"] = False
            retDict["Write"] = False
            retDict["Remove"] = False
            retDict["Check"] = False
            retDict["DiskSE"] = False
            retDict["TapeSE"] = False
            retDict["TotalCapacityTB"] = -1
            retDict["DiskCacheTB"] = -1
            return retDict

        # If nothing is defined in the CS Access is allowed
        # If something is defined, then it must be set to Active
        retDict["Read"] = not (
            "ReadAccess" in self.options and self.options["ReadAccess"] not in ("Active", "Degraded")
        )
        retDict["Write"] = not (
            "WriteAccess" in self.options and self.options["WriteAccess"] not in ("Active", "Degraded")
        )
        retDict["Remove"] = not (
            "RemoveAccess" in self.options and self.options["RemoveAccess"] not in ("Active", "Degraded")
        )
        if retDict["Read"]:
            retDict["Check"] = True
        else:
            retDict["Check"] = not (
                "CheckAccess" in self.options and self.options["CheckAccess"] not in ("Active", "Degraded")
            )
        diskSE = True
        tapeSE = False
        if "SEType" in self.options:
            # Type should follow the convention TXDY
            seType = self.options["SEType"]
            diskSE = re.search("D[1-9]", seType) is not None
            tapeSE = re.search("T[1-9]", seType) is not None
        retDict["DiskSE"] = diskSE
        retDict["TapeSE"] = tapeSE
        try:
            retDict["TotalCapacityTB"] = float(self.options["TotalCapacityTB"])
        except Exception:
            retDict["TotalCapacityTB"] = -1
        try:
            retDict["DiskCacheTB"] = float(self.options["DiskCacheTB"])
        except Exception:
            retDict["DiskCacheTB"] = -1

        return retDict

    def isValid(self, operation=None):
        """check CS/RSS statuses for :operation:

        :param str operation: operation name
        """
        log = self.log.getSubLogger("isValid")
        log.debug("Determining if the StorageElement %s is valid for VO %s" % (self.name, self.vo))

        if not self.valid:
            log.debug("Failed to create StorageElement plugins.", self.errorReason)
            return S_ERROR("SE.isValid: Failed to create StorageElement plugins: %s" % self.errorReason)

        # Check if the Storage Element is eligible for the user's VO
        if "VO" in self.options and self.vo not in self.options["VO"]:
            log.debug("StorageElement is not allowed for VO", self.vo)
            return S_ERROR(errno.EACCES, "StorageElement.isValid: StorageElement is not allowed for VO")
        log.debug("Determining if the StorageElement %s is valid for operation '%s'" % (self.name, operation))
        if (not operation) or (operation in self.okMethods):
            return S_OK()

        # Determine whether the StorageElement is valid for checking, reading, writing
        status = self.status()
        checking = status["Check"]
        reading = status["Read"]
        writing = status["Write"]
        removing = status["Remove"]

        # Determine whether the requested operation can be fulfilled
        if (not operation) and (not reading) and (not writing) and (not checking):
            log.debug("Read, write and check access not permitted.")
            return S_ERROR(errno.EACCES, "SE.isValid: Read, write and check access not permitted.")

        # The supplied operation can be 'Read','Write' or any of the possible StorageElement methods.
        if (
            (operation in self.readMethods)
            or (operation in self.stageMethods)
            or (operation.lower() in ("read", "readaccess"))
        ):
            operation = "ReadAccess"
        elif operation in self.writeMethods or (operation.lower() in ("write", "writeaccess")):
            operation = "WriteAccess"
        elif operation in self.removeMethods or (operation.lower() in ("remove", "removeaccess")):
            operation = "RemoveAccess"
        elif operation in self.checkMethods or (operation.lower() in ("check", "checkaccess")):
            operation = "CheckAccess"
        else:
            log.debug("The supplied operation is not known.", operation)
            return S_ERROR(DErrno.ENOMETH, "SE.isValid: The supplied operation is not known.")
        log.debug("check the operation: %s " % operation)

        # Check if the operation is valid
        if operation == "CheckAccess":
            if not reading:
                if not checking:
                    log.debug("Check access not currently permitted.")
                    return S_ERROR(errno.EACCES, "SE.isValid: Check access not currently permitted.")
        if operation == "ReadAccess":
            if not reading:
                log.debug("Read access not currently permitted.")
                return S_ERROR(errno.EACCES, "SE.isValid: Read access not currently permitted.")
        if operation == "WriteAccess":
            if not writing:
                log.debug("Write access not currently permitted.")
                return S_ERROR(errno.EACCES, "SE.isValid: Write access not currently permitted.")
        if operation == "RemoveAccess":
            if not removing:
                log.debug("Remove access not currently permitted.")
                return S_ERROR(errno.EACCES, "SE.isValid: Remove access not currently permitted.")
        return S_OK()

    def getPlugins(self):
        """Get the list of all the plugins defined for this Storage Element"""
        self.log.getSubLogger("getPlugins").debug("Obtaining all plugins of %s." % self.name)
        if not self.valid:
            return S_ERROR(self.errorReason)
        allPlugins = self.localPlugins + self.remotePlugins
        return S_OK(allPlugins)

    def getRemotePlugins(self):
        """Get the list of all the remote access protocols defined for this Storage Element"""
        self.log.getSubLogger("getRemotePlugins").debug("Obtaining remote protocols for %s." % self.name)
        if not self.valid:
            return S_ERROR(self.errorReason)
        return S_OK(self.remotePlugins)

    def getLocalPlugins(self):
        """Get the list of all the local access protocols defined for this Storage Element"""
        self.log.getSubLogger("getLocalPlugins").debug("Obtaining local protocols for %s." % self.name)
        if not self.valid:
            return S_ERROR(self.errorReason)
        return S_OK(self.localPlugins)

    def getStorageParameters(self, plugin=None, protocol=None):
        """Get plugin specific options

        :param plugin: plugin we are interested in
        :param protocol: protocol we are interested in

        Either plugin or protocol can be defined, not both, but at least one of them
        """

        # both set
        if plugin and protocol:
            return S_ERROR(errno.EINVAL, "plugin and protocol cannot be set together.")
        # both None
        elif not (plugin or protocol):
            return S_ERROR(errno.EINVAL, "plugin and protocol cannot be None together.")

        log = self.log.getSubLogger("getStorageParameters")

        reqStr = "plugin %s" % plugin if plugin else "protocol %s" % protocol

        log.debug("Obtaining storage parameters for %s for %s." % (self.name, reqStr))

        for storage in self.storages:
            storageParameters = storage.getParameters()
            if plugin and storageParameters["PluginName"] == plugin:
                return S_OK(storageParameters)
            elif protocol and storageParameters["Protocol"] == protocol:
                return S_OK(storageParameters)

        errStr = "Requested plugin or protocol not available."
        log.debug(errStr, "%s for %s" % (reqStr, self.name))
        return S_ERROR(errno.ENOPROTOOPT, errStr)

    def __getAllProtocols(self, protoType):
        """Returns the list of all protocols for Input or Output

        :param proto = InputProtocols or OutputProtocols

        """
        return set(reduce(lambda x, y: x + y, [plugin.protocolParameters[protoType] for plugin in self.storages]))

    def _getAllInputProtocols(self):
        """Returns all the protocols supported by the SE for Input"""
        return self.__getAllProtocols("InputProtocols")

    def _getAllOutputProtocols(self):
        """Returns all the protocols supported by the SE for Output"""
        return self.__getAllProtocols("OutputProtocols")

    def generateTransferURLsBetweenSEs(self, lfns, sourceSE, protocols=None):
        """This negotiate the URLs to be used for third party copy.
        This is mostly useful for FTS. If protocols is given,
        it restricts the list of plugins to use

        :param lfns: list/dict of lfns to generate the URLs
        :param sourceSE: storageElement instance of the sourceSE
        :param protocols: ordered protocol restriction list

        :return: dictionary with keys:

          * Successful: lfn indexed pair (src, dest) urls
          * Failed: lfn indexed with error
          * Protocols: tuple (srcProto, destProto)
        """
        log = self.log.getSubLogger("generateTransferURLsBetweenSEs")

        result = checkArgumentFormat(lfns)
        if result["OK"]:
            lfns = result["Value"]
        else:
            errStr = "Supplied urls must be string, list of strings or a dictionary."
            log.debug(errStr)
            return S_ERROR(errno.EINVAL, errStr)

        # First, find common protocols to use
        res = self.negociateProtocolWithOtherSE(sourceSE, protocols=protocols)

        if not res["OK"]:
            return res

        commonProtocols = res["Value"]

        # We sort the storage plugins based on their native protocol
        # according to the common protocol list
        # This is to favor for example the xroot plugin over the SRM plugin
        # even if both can provide xroot
        sourceSEStorages = sorted(
            sourceSE.storages, key=lambda x: getIndexInList(x.getParameters()["Protocol"], commonProtocols)
        )

        selfStorages = sorted(
            self.storages, key=lambda x: getIndexInList(x.getParameters()["Protocol"], commonProtocols)
        )

        # Taking each protocol at the time, we try to generate src and dest URLs
        for proto in commonProtocols:
            srcPlugin = None
            destPlugin = None

            log.debug("Trying to find plugins for protocol %s" % proto)

            # Finding the source storage plugin
            for storagePlugin in sourceSEStorages:
                log.debug("Testing %s as source plugin" % storagePlugin.pluginName)
                storageParameters = storagePlugin.getParameters()
                nativeProtocol = storageParameters["Protocol"]
                # If the native protocol of the plugin is allowed for read
                if nativeProtocol in sourceSE.localAccessProtocolList:
                    # If the plugin can generate the protocol we are interested in
                    if proto in storageParameters["OutputProtocols"]:
                        log.debug("Selecting it")
                        srcPlugin = storagePlugin
                        break
            # If we did not find a source plugin, continue
            if srcPlugin is None:
                log.debug("Could not find a source plugin for protocol %s" % proto)
                continue

            # Finding the destination storage plugin
            for storagePlugin in selfStorages:
                log.debug("Testing %s as destination plugin" % storagePlugin.pluginName)

                storageParameters = storagePlugin.getParameters()
                nativeProtocol = storageParameters["Protocol"]
                # If the native protocol of the plugin is allowed for write
                if nativeProtocol in self.localWriteProtocolList:
                    # If the plugin can accept the protocol we are interested in
                    if proto in storageParameters["InputProtocols"]:
                        log.debug("Selecting it")
                        destPlugin = storagePlugin
                        break

            # If we found both a source and destination plugin, we are happy,
            # otherwise we continue with the next protocol
            if destPlugin is None:
                log.debug("Could not find a destination plugin for protocol %s" % proto)
                srcPlugin = None
                continue

            failed = {}
            successful = {}
            # Generate the URLs
            for lfn in lfns:

                # Source URL first
                res = srcPlugin.constructURLFromLFN(lfn, withWSUrl=True)
                if not res["OK"]:
                    errMsg = "Error generating source url: %s" % res["Message"]
                    log.debug("Error generating source url", errMsg)
                    failed[lfn] = errMsg
                    continue
                srcURL = res["Value"]

                # Destination URL
                res = destPlugin.constructURLFromLFN(lfn, withWSUrl=True)
                if not res["OK"]:
                    errMsg = "Error generating destination url: %s" % res["Message"]
                    log.debug("Error generating destination url", errMsg)
                    failed[lfn] = errMsg
                    continue
                destURL = res["Value"]

                successful[lfn] = (srcURL, destURL)

            nativeSrcProtocol = srcPlugin.getParameters()["Protocol"]
            nativeDestProtocol = destPlugin.getParameters()["Protocol"]
            return S_OK(
                {"Successful": successful, "Failed": failed, "Protocols": (nativeSrcProtocol, nativeDestProtocol)}
            )

        return S_ERROR(errno.ENOPROTOOPT, "Could not find a protocol ")

    def negociateProtocolWithOtherSE(self, sourceSE, protocols=None):
        """Negotiate what protocol could be used for a third party transfer
        between the sourceSE and ourselves. If protocols is given,
        the chosen protocol has to be among those

        :param sourceSE: storageElement instance of the sourceSE
        :param protocols: ordered protocol restriction list

        :return: a list protocols that fits the needs, or None

        """

        # No common protocols if this is a proxy storage
        if self.useProxy:
            return S_OK([])

        log = self.log.getSubLogger("negociateProtocolWithOtherSE")

        log.debug("Negotiating protocols between %s and %s (protocols %s)" % (sourceSE.name, self.name, protocols))

        # Take all the protocols the destination can accept as input
        destProtocols = self._getAllInputProtocols()

        log.debug("Destination input protocols %s" % destProtocols)

        # Take all the protocols the source can provide
        sourceProtocols = sourceSE._getAllOutputProtocols()

        log.debug("Source output protocols %s" % sourceProtocols)

        commonProtocols = destProtocols & sourceProtocols

        # If a restriction list is defined
        # take the intersection, and sort the commonProtocols
        # based on the protocolList order
        if protocols:
            protocolList = list(protocols)
            commonProtocols = sorted(commonProtocols & set(protocolList), key=lambda x: getIndexInList(x, protocolList))

        log.debug("Common protocols %s" % commonProtocols)

        return S_OK(list(commonProtocols))

    #################################################################################################
    #
    # These are the basic get functions for lfn manipulation
    #

    def __getURLPath(self, url):
        """Get the part of the URL path below the basic storage path.
        This path must coincide with the LFN of the file in order to be compliant with the DIRAC conventions.
        """
        log = self.log.getSubLogger("__getURLPath")
        log.debug("Getting path from url in %s." % self.name)
        if not self.valid:
            return S_ERROR(self.errorReason)
        res = pfnparse(url)
        if not res["OK"]:
            return res
        fullURLPath = "%s/%s" % (res["Value"]["Path"], res["Value"]["FileName"])

        # Check all available storages and check whether the url is for that protocol
        urlPath = ""
        for storage in self.storages:
            res = storage.isNativeURL(url)
            if res["OK"]:
                if res["Value"]:
                    parameters = storage.getParameters()
                    saPath = parameters["Path"]
                    if not saPath:
                        # If the sa path doesn't exist then the url path is the entire string
                        urlPath = fullURLPath
                    else:
                        if re.search(saPath, fullURLPath):
                            # Remove the sa path from the fullURLPath
                            urlPath = fullURLPath.replace(saPath, "")
            if urlPath:
                return S_OK(urlPath)
        # This should never happen. DANGER!!
        errStr = "Failed to get the url path for any of the protocols!!"
        log.debug(errStr)
        return S_ERROR(errStr)

    def getLFNFromURL(self, urls):
        """Get the LFN from the PFNS .

        :param lfn: input lfn or lfns (list/dict)

        """
        result = checkArgumentFormat(urls)
        if result["OK"]:
            urlDict = result["Value"]
        else:
            errStr = "Supplied urls must be string, list of strings or a dictionary."
            self.log.getSubLogger("getLFNFromURL").debug(errStr)
            return S_ERROR(errno.EINVAL, errStr)

        retDict = {"Successful": {}, "Failed": {}}
        for url in urlDict:
            res = self.__getURLPath(url)
            if res["OK"]:
                retDict["Successful"][url] = res["Value"]
            else:
                retDict["Failed"][url] = res["Message"]
        return S_OK(retDict)

    ###########################################################################################
    #
    # This is the generic wrapper for file operations
    #

    def getURL(self, lfn, protocol=False, replicaDict=None):
        """execute 'getTransportURL' operation.

        :param str lfn: string, list or dictionary of lfns
        :param protocol: if no protocol is specified, we will request self.turlProtocols
        :param replicaDict: optional results from the File Catalog replica query

        """

        self.log.getSubLogger("getURL").debug(
            "Getting accessUrl %s for lfn in %s." % ("(%s)" % protocol if protocol else "", self.name)
        )

        if not protocol:
            # This turlProtocols seems totally useless.
            # Get ride of it when gfal2 is totally ready
            # and replace it with the localAccessProtocol list
            protocols = self.turlProtocols
        elif isinstance(protocol, list):
            protocols = protocol
        elif isinstance(protocol, six.string_types):
            protocols = [protocol]

        self.methodName = "getTransportURL"
        result = self.__executeMethod(lfn, protocols=protocols)
        return result

    def __isLocalSE(self):
        """Test if the Storage Element is local in the current context"""
        self.log.getSubLogger("LocalSE").debug("Determining whether %s is a local SE." % self.name)

        import DIRAC

        localSEs = getSEsForSite(DIRAC.siteName())["Value"]
        if self.name in localSEs:
            return S_OK(True)
        else:
            return S_OK(False)

    def __getFileCatalog(self):

        if not self.__fileCatalog:
            self.__fileCatalog = FileCatalog(vo=self.vo)
        return self.__fileCatalog

    def __generateURLDict(self, lfns, storage, replicaDict=None):
        """Generates a dictionary (url : lfn ), where the url are constructed
        from the lfn using the constructURLFromLFN method of the storage plugins.

        :param lfns: dictionary {lfn:whatever}

        :returns: dictionary {constructed url : lfn}
        """
        log = self.log.getSubLogger("__generateURLDict")
        log.debug("generating url dict for %s lfn in %s." % (len(lfns), self.name))

        if not replicaDict:
            replicaDict = {}

        urlDict = {}  # url : lfn
        failed = {}  # lfn : string with errors
        for lfn in lfns:
            if self.useCatalogURL:
                # Is this self.name alias proof?
                url = replicaDict.get(lfn, {}).get(self.name, "")
                if url:
                    urlDict[url] = lfn
                    continue
                else:
                    fc = self.__getFileCatalog()
                    result = fc.getReplicas()
                    if not result["OK"]:
                        failed[lfn] = result["Message"]
                    url = result["Value"]["Successful"].get(lfn, {}).get(self.name, "")

                if not url:
                    failed[lfn] = "Failed to get catalog replica"
                else:
                    # Update the URL according to the current SE description
                    result = returnSingleResult(storage.updateURL(url))
                    if not result["OK"]:
                        failed[lfn] = result["Message"]
                    else:
                        urlDict[result["Value"]] = lfn
            else:
                result = storage.constructURLFromLFN(lfn, withWSUrl=True)
                if not result["OK"]:
                    errStr = result["Message"]
                    log.debug(errStr, "for %s" % (lfn))
                    failed[lfn] = "%s %s" % (failed[lfn], errStr) if lfn in failed else errStr
                else:
                    urlDict[result["Value"]] = lfn

        res = S_OK({"Successful": urlDict, "Failed": failed})
        #     res['Failed'] = failed
        return res

    def __filterPlugins(self, methodName, protocols=None, inputProtocol=None):
        """Determine the list of plugins that
         can be used for a particular action

         Args:
           method(str): method to execute
           protocols(list): specific protocols might be requested
           inputProtocol(str): in case the method is putFile, this specifies
                               the protocol given as source

        Returns:
          list: list of storage plugins
        """

        log = self.log.getSubLogger("__filterPlugins")

        log.debug(
            "Filtering plugins for %s (protocol = %s ; inputProtocol = %s)" % (methodName, protocols, inputProtocol)
        )

        if isinstance(protocols, six.string_types):
            protocols = [protocols]

        pluginsToUse = []

        potentialProtocols = []
        allowedProtocols = []

        if methodName in self.readMethods + self.checkMethods:
            allowedProtocols = self.localAccessProtocolList
        elif methodName in self.stageMethods:
            allowedProtocols = self.localStageProtocolList
        elif methodName in self.removeMethods + self.writeMethods:
            allowedProtocols = self.localWriteProtocolList
        else:
            # OK methods
            # If a protocol or protocol list is specified, we only use the plugins that
            # can generate such protocol
            # otherwise we return them all
            if protocols:
                setProtocol = set(protocols)
                for plugin in self.storages:
                    if set(plugin.protocolParameters.get("OutputProtocols", [])) & setProtocol:
                        log.debug("Plugin %s can generate compatible protocol" % plugin.pluginName)
                        pluginsToUse.append(plugin)
            else:
                pluginsToUse = self.storages

            # The closest list for "OK" methods is the AccessProtocol preference, so we sort based on that
            pluginsToUse.sort(
                key=lambda x: getIndexInList(x.protocolParameters["Protocol"], self.localAccessProtocolList)
            )
            log.debug("Plugins to be used for %s: %s" % (methodName, [p.pluginName for p in pluginsToUse]))
            return pluginsToUse

        log.debug("Allowed protocol: %s" % allowedProtocols)

        # if a list of protocol is specified, take it into account
        if protocols:
            potentialProtocols = list(set(allowedProtocols) & set(protocols))
        else:
            potentialProtocols = allowedProtocols

        log.debug("Potential protocols %s" % potentialProtocols)

        localSE = self.__isLocalSE()["Value"]

        for plugin in self.storages:
            # Determine whether to use this storage object
            pluginParameters = plugin.getParameters()
            pluginName = pluginParameters.get("PluginName")

            if not pluginParameters:
                log.debug("Failed to get storage parameters.", "%s %s" % (self.name, pluginName))
                continue

            if not (pluginName in self.remotePlugins) and not localSE and not pluginName == "Proxy":
                # If the SE is not local then we can't use local protocols
                log.debug("Local protocol not appropriate for remote use: %s." % pluginName)
                continue

            if pluginParameters["Protocol"] not in potentialProtocols:
                log.debug("Plugin %s not allowed for %s." % (pluginName, methodName))
                continue

            # If we are attempting a putFile and we know the inputProtocol
            if methodName == "putFile" and inputProtocol:
                if inputProtocol not in pluginParameters["InputProtocols"]:
                    log.debug("Plugin %s not appropriate for %s protocol as input." % (pluginName, inputProtocol))
                    continue

            pluginsToUse.append(plugin)

        # sort the plugins according to the lists in the CS
        pluginsToUse.sort(key=lambda x: getIndexInList(x.protocolParameters["Protocol"], allowedProtocols))

        log.debug("Plugins to be used for %s: %s" % (methodName, [p.pluginName for p in pluginsToUse]))

        return pluginsToUse

    def __executeMethod(self, lfn, *args, **kwargs):
        """Forward the call to each storage in turn until one works.
        The method to be executed is stored in self.methodName

        :param lfn: string, list or dictionary
        :param *args: variable amount of non-keyword arguments. SHOULD BE EMPTY
        :param **kwargs: keyword arguments

        A special argument is 'protocols', which will be used by the StorageElement to filter
        the usable plugins. Unless the method being executed is getTransportURL, this parameter
        is removed from kwargs.
        A special kwargs is 'inputProtocol', which can be specified for putFile. It describes
        the protocol used as source protocol, since there is in principle only one.


        :returns: S_OK( { 'Failed': {lfn : reason} , 'Successful': {lfn : value} } )
                The Failed dict contains the lfn only if the operation failed on all the storages
                The Successful dict contains the value returned by the successful storages.

        """

        removedArgs = {}
        log = self.log.getSubLogger("__executeMethod")
        log.debug("preparing the execution of %s" % (self.methodName))

        # args should normaly be empty to avoid problem...
        if args:
            log.debug("args should be empty!%s" % args)
            # because there is normally only one kw argument, I can move it from args to kwargs
            methDefaultArgs = list(StorageElementItem.__defaultsArguments.get(self.methodName, {}))
            if methDefaultArgs:
                kwargs[methDefaultArgs[0]] = args[0]
                args = args[1:]
            log.debug("put it in kwargs, but dirty and might be dangerous!args %s kwargs %s" % (args, kwargs))

        # We check the deprecated arguments
        for depArg in StorageElementItem.__deprecatedArguments:
            if depArg in kwargs:
                log.verbose("%s is not an allowed argument anymore. Please change your code!" % depArg)
                removedArgs[depArg] = kwargs[depArg]
                del kwargs[depArg]

        # Set default argument if any
        methDefaultArgs = StorageElementItem.__defaultsArguments.get(self.methodName, {})
        for argName in methDefaultArgs:
            if argName not in kwargs:
                log.debug(
                    "default argument %s for %s not present.\
         Setting value %s."
                    % (argName, self.methodName, methDefaultArgs[argName])
                )
                kwargs[argName] = methDefaultArgs[argName]

        res = checkArgumentFormat(lfn)
        if not res["OK"]:
            errStr = "Supplied lfns must be string, list of strings or a dictionary."
            log.debug(errStr)
            return res
        lfnDict = res["Value"]

        log.debug("Attempting to perform '%s' operation with %s lfns." % (self.methodName, len(lfnDict)))

        res = self.isValid(operation=self.methodName)
        if not res["OK"]:
            return res
        else:
            if not self.valid:
                return S_ERROR(self.errorReason)
        # In case executing putFile, we can assume that all the source urls
        # are from the same protocol. This optional parameter, if defined
        # can be used to ignore some storage plugins and thus save time
        # and avoid fake failures showing in the accounting
        inputProtocol = kwargs.pop("inputProtocol", None)

        # the 'protocols' parameter is only given to the plugin when calling getTransportURL.
        # The other methods do not expect it.
        protocols = kwargs.get("protocols")
        if self.methodName != "getTransportURL":
            kwargs.pop("protocols", None)

        successful = {}
        failed = {}
        filteredPlugins = self.__filterPlugins(self.methodName, protocols, inputProtocol)
        if not filteredPlugins:
            return S_ERROR(
                errno.EPROTONOSUPPORT,
                "No storage plugins matching the requirements\
                                           (operation %s protocols %s inputProtocol %s)"
                % (self.methodName, protocols, inputProtocol),
            )
        # Try all of the storages one by one
        for storage in filteredPlugins:
            # Determine whether to use this storage object
            storageParameters = storage.getParameters()
            pluginName = storageParameters["PluginName"]

            if not lfnDict:
                log.debug("No lfns to be attempted for %s protocol." % pluginName)
                continue

            log.debug("Generating %s protocol URLs for %s." % (len(lfnDict), pluginName))
            replicaDict = kwargs.pop("replicaDict", {})
            if storage.pluginName != "Proxy":
                res = self.__generateURLDict(lfnDict, storage, replicaDict=replicaDict)
                urlDict = res["Value"]["Successful"]  # url : lfn
                failed.update(res["Value"]["Failed"])
            else:
                urlDict = dict([(lfn, lfn) for lfn in lfnDict])
            if not urlDict:
                log.debug("__executeMethod No urls generated for protocol %s." % pluginName)
            else:
                log.debug("Attempting to perform '%s' for %s physical files" % (self.methodName, len(urlDict)))
                fcn = None
                if hasattr(storage, self.methodName) and callable(getattr(storage, self.methodName)):
                    fcn = getattr(storage, self.methodName)
                if not fcn:
                    return S_ERROR(
                        DErrno.ENOMETH, "SE.__executeMethod: unable to invoke %s, it isn't a member function of storage"
                    )
                urlsToUse = {}  # url : the value of the lfn dictionary for the lfn of this url
                for url in urlDict:
                    urlsToUse[url] = lfnDict[urlDict[url]]

                startDate = datetime.datetime.utcnow()
                startTime = time.time()
                res = fcn(urlsToUse, *args, **kwargs)
                elapsedTime = time.time() - startTime

                self.addAccountingOperation(urlsToUse, startDate, elapsedTime, storageParameters, res)

                if not res["OK"]:
                    errStr = "Completely failed to perform %s." % self.methodName
                    log.verbose(errStr, "with plugin %s: %s" % (pluginName, res["Message"]))
                    for lfn in urlDict.values():
                        if lfn not in failed:
                            failed[lfn] = ""
                        failed[lfn] = "%s %s" % (failed[lfn], res["Message"]) if failed[lfn] else res["Message"]

                else:
                    for url, lfn in urlDict.items():
                        if url not in res["Value"]["Successful"]:
                            if lfn not in failed:
                                failed[lfn] = ""
                            if url in res["Value"]["Failed"]:
                                self.log.verbose(
                                    "Failure in plugin to perform %s" % self.methodName,
                                    "Plugin: %s lfn: %s error %s" % (pluginName, lfn, res["Value"]["Failed"][url]),
                                )
                                failed[lfn] = (
                                    "%s %s" % (failed[lfn], res["Value"]["Failed"][url])
                                    if failed[lfn]
                                    else res["Value"]["Failed"][url]
                                )
                            else:
                                errStr = "No error returned from plug-in"
                                failed[lfn] = "%s %s" % (failed[lfn], errStr) if failed[lfn] else errStr
                        else:
                            successful[lfn] = res["Value"]["Successful"][url]
                            if lfn in failed:
                                failed.pop(lfn)
                            lfnDict.pop(lfn)

        gDataStoreClient.commit()

        return S_OK({"Failed": failed, "Successful": successful})

    def __getattr__(self, name):
        """Forwards the equivalent Storage calls to __executeMethod"""
        # We take either the equivalent name, or the name itself
        self.methodName = StorageElementItem.__equivalentMethodNames.get(name, None)

        if self.methodName:
            return self.__executeMethod

        raise AttributeError("StorageElement does not have a method '%s'" % name)

    def addAccountingOperation(self, lfns, startDate, elapsedTime, storageParameters, callRes):
        """
        Generates a DataOperation accounting if needs to be, and adds it to the DataStore client cache

        :param lfns: list of lfns on which we attempted the operation
        :param startDate: datetime, start of the operation
        :param elapsedTime: time (seconds) the operation took
        :param storageParameters: the parameters of the plugins used to perform the operation
        :param callRes: the return of the method call, S_OK or S_ERROR

        The operation is generated with the OperationType "se.methodName"
        The TransferSize and TransferTotal for directory methods actually take into
        account the files inside the directory, and not the amount of directory given
        as parameter


        """

        if self.methodName not in (self.readMethods + self.writeMethods + self.removeMethods + self.stageMethods):
            return

        baseAccountingDict = {}
        baseAccountingDict["OperationType"] = "se.%s" % self.methodName
        baseAccountingDict["User"] = getProxyInfo().get("Value", {}).get("username", "unknown")
        baseAccountingDict["RegistrationTime"] = 0.0
        baseAccountingDict["RegistrationOK"] = 0
        baseAccountingDict["RegistrationTotal"] = 0

        # if it is a get method, then source and destination of the transfer should be inverted
        if self.methodName == "getFile":
            baseAccountingDict["Destination"] = siteName()
            baseAccountingDict["Source"] = self.name
        else:
            baseAccountingDict["Destination"] = self.name
            baseAccountingDict["Source"] = siteName()

        baseAccountingDict["TransferTotal"] = 0
        baseAccountingDict["TransferOK"] = 0
        baseAccountingDict["TransferSize"] = 0
        baseAccountingDict["TransferTime"] = 0.0
        baseAccountingDict["FinalStatus"] = "Successful"

        oDataOperation = DataOperation()
        oDataOperation.setValuesFromDict(baseAccountingDict)
        oDataOperation.setStartTime(startDate)
        oDataOperation.setEndTime(startDate + datetime.timedelta(seconds=elapsedTime))
        oDataOperation.setValueByKey("TransferTime", elapsedTime)
        oDataOperation.setValueByKey("Protocol", storageParameters.get("Protocol", "unknown"))

        if not callRes["OK"]:
            # Everything failed
            oDataOperation.setValueByKey("TransferTotal", len(lfns))
            oDataOperation.setValueByKey("FinalStatus", "Failed")
        else:

            succ = callRes.get("Value", {}).get("Successful", {})
            failed = callRes.get("Value", {}).get("Failed", {})

            totalSize = 0
            # We don't take len(lfns) in order to make two
            # separate entries in case of few failures
            totalSucc = len(succ)

            if self.methodName in ("putFile", "getFile"):
                # putFile and getFile return for each entry
                # in the successful dir the size of the corresponding file
                totalSize = sum(succ.values())

            elif self.methodName in ("putDirectory", "getDirectory"):
                # putDirectory and getDirectory return for each dir name
                # a dictionnary with the keys 'Files' and 'Size'
                totalSize = sum(val.get("Size", 0) for val in succ.values() if isinstance(val, dict))
                totalSucc = sum(val.get("Files", 0) for val in succ.values() if isinstance(val, dict))
                oDataOperation.setValueByKey("TransferOK", len(succ))

            oDataOperation.setValueByKey("TransferSize", totalSize)
            oDataOperation.setValueByKey("TransferTotal", totalSucc)
            oDataOperation.setValueByKey("TransferOK", totalSucc)

            if callRes["Value"]["Failed"]:
                oDataOperationFailed = copy.deepcopy(oDataOperation)
                oDataOperationFailed.setValueByKey("TransferTotal", len(failed))
                oDataOperationFailed.setValueByKey("TransferOK", 0)
                oDataOperationFailed.setValueByKey("TransferSize", 0)
                oDataOperationFailed.setValueByKey("FinalStatus", "Failed")

                accRes = gDataStoreClient.addRegister(oDataOperationFailed)
                if not accRes["OK"]:
                    self.log.error("Could not send failed accounting report", accRes["Message"])

        accRes = gDataStoreClient.addRegister(oDataOperation)
        if not accRes["OK"]:
            self.log.error("Could not send accounting report", accRes["Message"])


StorageElement = StorageElementCache()



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-lfn-replicas
# Author :  Stuart Paterson
########################################################################
"""
Obtain replica information from file catalogue client.

Example:
  $ dirac-dms-lfn-replicas /formation/user/v/vhamar/Test.txt
  {'Failed': {},
   'Successful': {'/formation/user/v/vhamar/Test.txt':\
   {'M3PEC-disk': 'srm://se0.m3pec.u-bordeaux1.fr/dpm/m3pec.u-bordeaux1.fr/home/formation/user/v/vhamar/Test.txt'}}}
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("a", "All", "  Also show inactive replicas")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["LFN:      Logical File Name or file containing LFNs"])
    switches, lfns = Script.parseCommandLine(ignoreErrors=True)

    active = True
    for switch in switches:
        opt = switch[0].lower()
        if opt in ("a", "all"):
            active = False

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    exitCode = 0

    if len(lfns) == 1:
        try:
            with open(lfns[0], "r") as f:
                lfns = f.read().splitlines()
        except Exception:
            pass

    result = dirac.getReplicas(lfns, active=active, printOutput=True)
    if not result["OK"]:
        print("ERROR: ", result["Message"])
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



########################################################################
# $HeadURL $
# File: GraphTests.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2012/09/28 09:02:23
########################################################################
""" :mod: GraphTests
    =======================

    .. module: GraphTests
    :synopsis: tests for Graph module classes
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
# #
# @file GraphTests.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2012/09/28 09:02:24
# @brief Definition of GraphTests class.
# # imports
import unittest
import six

# # SUT
from DIRAC.Core.Utilities.Graph import Node, Edge, Graph, DynamicProps  # , topologicalSort, topoSort


class DynamicPropTests(unittest.TestCase):
    """
    ..  class:: DynamicPropTests
    """

    def testDynamicProps(self):
        """test dynamic props"""

        @six.add_metaclass(DynamicProps)
        class TestClass(object):
            """
            .. class:: TestClass

            dummy class
            """

        # # dummy instance
        testObj = TestClass()
        # # makeProperty in
        self.assertEqual(hasattr(testObj, "makeProperty"), True)
        self.assertEqual(callable(getattr(testObj, "makeProperty")), True)
        # # .. and works  for rw properties
        testObj.makeProperty("rwTestProp", 10)  # pylint: disable=no-member
        self.assertEqual(hasattr(testObj, "rwTestProp"), True)
        self.assertEqual(getattr(testObj, "rwTestProp"), 10)
        testObj.rwTestProp += 1  # pylint: disable=no-member
        self.assertEqual(getattr(testObj, "rwTestProp"), 11)
        # # .. and ro as well
        testObj.makeProperty("roTestProp", "I'm read only", True)  # pylint: disable=no-member
        self.assertEqual(hasattr(testObj, "roTestProp"), True)
        self.assertEqual(getattr(testObj, "roTestProp"), "I'm read only")
        # # AttributeError for read only property setattr
        try:
            testObj.roTestProp = 11
        except AttributeError as error:
            self.assertEqual(str(error), "can't set attribute")


class NodeTests(unittest.TestCase):
    """
    .. class:: NodeTests
    """

    def setUp(self):
        """test setup"""
        self.roAttrs = {"ro1": True, "ro2": "I'm read only"}
        self.rwAttrs = {"rw1": 0, "rw2": (1, 2, 3)}
        self.name = "BrightStart"
        self.node = Node(self.name, self.rwAttrs, self.roAttrs)

    def tearDown(self):
        """clean up"""
        del self.roAttrs
        del self.rwAttrs
        del self.name
        del self.node

    def testNode(self):
        """node rwAttrs roAttrs connect"""

        # # node name - th eon,y one prop you can't overwrite
        self.assertEqual(self.node.name, self.name)
        try:
            self.node.name = "can't do this"
        except AttributeError as error:
            self.assertEqual(str(error), "can't set attribute")
        try:
            self.node.makeProperty("name", "impossible")
        except AttributeError as error:
            self.assertEqual(str(error), "_name or name is already defined as a member")

        # # visited attr for walking
        self.assertEqual(hasattr(self.node, "visited"), True)
        self.assertEqual(self.node.visited, False)  # pylint: disable=no-member

        # # ro attrs
        for k, v in self.roAttrs.items():
            self.assertEqual(hasattr(self.node, k), True)
            self.assertEqual(getattr(self.node, k), v)
            try:
                setattr(self.node, k, "new value")
            except AttributeError as error:
                self.assertEqual(str(error), "can't set attribute")

        # # rw attrs
        for k, v in self.rwAttrs.items():
            self.assertEqual(hasattr(self.node, k), True)
            self.assertEqual(getattr(self.node, k), v)
            setattr(self.node, k, "new value")
            self.assertEqual(getattr(self.node, k), "new value")

        # # connect
        toNode = Node("DeadEnd")
        edge = self.node.connect(toNode, {"foo": "boo"}, {"ro3": True})
        self.assertEqual(isinstance(edge, Edge), True)
        self.assertEqual(edge.name, self.name + "-DeadEnd")
        self.assertEqual(self.node, edge.fromNode)  # pylint: disable=no-member
        self.assertEqual(toNode, edge.toNode)  # pylint: disable=no-member


class EdgeTests(unittest.TestCase):
    """
    .. class:: EdgeTests
    """

    def setUp(self):
        """test setup"""
        self.fromNode = Node("Start")
        self.toNode = Node("End")
        self.roAttrs = {"ro1": True, "ro2": "I'm read only"}
        self.rwAttrs = {"rw1": 0, "rw2": (1, 2, 3)}

    def tearDown(self):
        """clean up"""
        del self.fromNode
        del self.toNode
        del self.roAttrs
        del self.rwAttrs

    def testEdge(self):
        """c'tor connect attrs"""
        edge = Edge(self.fromNode, self.toNode, self.rwAttrs, self.roAttrs)

        # # name
        self.assertEqual(edge.name, "%s-%s" % (self.fromNode.name, self.toNode.name))
        try:
            edge.name = "can't do this"
        except AttributeError as error:
            self.assertEqual(str(error), "can't set attribute")
        try:
            edge.makeProperty("name", "impossible")
        except AttributeError as error:
            self.assertEqual(str(error), "_name or name is already defined as a member")

        # # visited attr
        self.assertEqual(hasattr(edge, "visited"), True)
        self.assertEqual(edge.visited, False)  # pylint: disable=no-member

        # # ro attrs
        for k, v in self.roAttrs.items():
            self.assertEqual(hasattr(edge, k), True)
            self.assertEqual(getattr(edge, k), v)
            try:
                setattr(edge, k, "new value")
            except AttributeError as error:
                self.assertEqual(str(error), "can't set attribute")

        # # rw attrs
        for k, v in self.rwAttrs.items():
            self.assertEqual(hasattr(edge, k), True)
            self.assertEqual(getattr(edge, k), v)
            setattr(edge, k, "new value")
            self.assertEqual(getattr(edge, k), "new value")

        # # start and end
        self.assertEqual(edge.fromNode, self.fromNode)  # pylint: disable=no-member
        self.assertEqual(edge.toNode, self.toNode)  # pylint: disable=no-member
        # # in fromNode, not in toNode
        self.assertEqual(edge in self.fromNode, True)
        self.assertEqual(edge not in self.toNode, True)


clock = 0
########################################################################
class GraphTests(unittest.TestCase):
    """
    .. class:: GraphTests
    """

    def setUp(self):
        """setup test case"""
        self.nodes = [Node("1"), Node("2"), Node("3")]
        self.edges = [self.nodes[0].connect(self.nodes[1]), self.nodes[0].connect(self.nodes[2])]
        self.aloneNode = Node("4")

    def tearDown(self):
        """clean up"""
        del self.nodes
        del self.edges
        del self.aloneNode

    def testGraph(self):
        """ctor nodes edges connect walk"""

        # # create graph
        gr = Graph("testGraph", self.nodes, self.edges)

        # # nodes and edges
        for node in self.nodes:
            self.assertEqual(node in gr, True)
        for edge in self.edges:
            self.assertEqual(edge in gr, True)
        self.assertEqual(sorted(self.nodes, key=lambda x: x.name), sorted(gr.nodes(), key=lambda x: x.name))
        self.assertEqual(sorted(self.edges, key=lambda x: x.name), sorted(gr.edges(), key=lambda x: x.name))

        # # getNode
        for node in self.nodes:
            self.assertEqual(gr.getNode(node.name), node)

        # # connect
        aloneEdge = gr.connect(self.nodes[0], self.aloneNode)
        self.assertEqual(self.aloneNode in gr, True)
        self.assertEqual(aloneEdge in gr, True)

        # # addNode
        anotherNode = Node("5")
        anotherEdge = anotherNode.connect(self.aloneNode)
        gr.addNode(anotherNode)
        self.assertEqual(anotherNode in gr, True)
        self.assertEqual(anotherEdge in gr, True)

        # # walk no nodeFcn
        ret = gr.walkAll()
        self.assertEqual(ret, {})

        for node in gr.nodes():
            self.assertEqual(node.visited, True)

        gr.reset()
        for node in gr.nodes():
            self.assertEqual(node.visited, False)
        # # walk with nodeFcn
        def nbEdges(node):
            """dummy node fcn"""
            return len(node.edges())

        ret = gr.walkAll(nodeFcn=nbEdges)
        self.assertEqual(ret, {"1": 3, "2": 0, "3": 0, "4": 0, "5": 1})

    def testDFS(self):
        """dfs"""

        global clock

        def topoA(graph):
            """topological sort"""
            global clock
            nodes = graph.nodes()
            for node in nodes:
                node.makeProperty("clockA", 0)

            def postVisit(node):
                global clock
                node.clockA = clock
                clock += 1

            graph.dfs(postVisit=postVisit)
            nodes = graph.nodes()
            nodes.sort(key=lambda node: node.clockA)
            return nodes

        def topoB(graph):
            """topological sort"""
            global clock
            nodes = graph.nodes()
            for node in nodes:
                node.makeProperty("clockB", 0)

            def postVisit(node):
                global clock
                node.clockB = clock
                clock += 1

            graph.dfsIter(postVisit=postVisit)
            nodes = graph.nodes()
            nodes.sort(key=lambda node: node.clockB)
            return nodes

        clock = 0
        gr = Graph("testGraph", self.nodes, self.edges)
        gr.addNode(self.aloneNode)
        nodesSorted = topoA(gr)
        nodes = gr.nodes()
        nodes.sort(key=lambda node: node.clockA, reverse=True)
        self.assertEqual(nodes, nodesSorted, "topoA sort failed")

        clock = 0
        gr = Graph("testGraph", self.nodes, self.edges)
        gr.addNode(self.aloneNode)
        gr.reset()
        nodesSorted = topoB(gr)
        nodes = gr.nodes()
        nodes.sort(key=lambda node: node.clockB, reverse=True)
        self.assertEqual(nodes, nodesSorted, "topoB sort failed")

    def testBFS(self):
        """bfs walk"""

        global clock

        def walk(graph):
            """bfs walk"""
            global clock
            nodes = graph.nodes()
            for node in nodes:
                node.makeProperty("clockC", 0)

            def postVisit(node):
                global clock
                node.clockC = clock
                clock += 1

            nodes = graph.bfs(postVisit=postVisit)
            nodes.sort(key=lambda node: node.clockC)
            return nodes

        clock = 0
        gr = Graph("testGraph", self.nodes, self.edges)
        gr.addNode(self.aloneNode)
        gr.reset()
        nodesSorted = walk(gr)
        nodes = gr.nodes()
        nodes.sort(key=lambda node: node.clockC)
        self.assertEqual(nodesSorted, nodes, "bfs failed")


# # test execution
if __name__ == "__main__":
    testLoader = unittest.TestLoader()
    tests = (
        testLoader.loadTestsFromTestCase(testCase) for testCase in (DynamicPropTests, NodeTests, EdgeTests, GraphTests)
    )
    testSuite = unittest.TestSuite(tests)
    unittest.TextTestRunner(verbosity=3).run(testSuite)



"""Test the JobInfo"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import sys
from six import StringIO

from mock import MagicMock as Mock

from parameterized import parameterized, param

from DIRAC import S_OK, S_ERROR, gLogger
import DIRAC

from DIRAC.TransformationSystem.Utilities.JobInfo import TaskInfoException, JobInfo
import DIRAC.Interfaces.API.Dirac

gLogger.setLevel("DEBUG")

__RCSID__ = "$Id$"

# pylint: disable=W0212, E1101


class TestJI(unittest.TestCase):
    """Test the JobInfo Module"""

    def setUp(self):
        self.jbi = JobInfo(jobID=123, status="Failed", tID=1234, tType="MCReconstruction")
        self.diracAPI = Mock(name="dilcMock", spec=DIRAC.Interfaces.API.Dirac.Dirac)
        self.jobMon = Mock(
            name="jobMonMock", spec=DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient.JobMonitoringClient
        )
        self.jobMon.getInputData = Mock(return_value=S_OK([]))
        self.jobMon.getJobAttribute = Mock(return_value=S_OK("0"))
        self.jobMon.getJobParameter = Mock(return_value=S_OK({}))
        self.diracAPI.getJobJDL = Mock()

        self.jdl2 = {
            "LogTargetPath": "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/LOG/00006326_015.tar",
            "Executable": "dirac-jobexec",
            "TaskID": 15,
            "SoftwareDistModule": "ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation",
            "JobName": "00006326_00000015",
            "Priority": 1,
            "Platform": "x86_64-slc5-gcc43-opt",
            "JobRequirements": {
                "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
                "VirtualOrganization": "ilc",
                "Setup": "ILC-Production",
                "CPUTime": 300000,
                "OwnerGroup": "ilc_prod",
                "Platforms": [
                    "x86_64-slc6-gcc44-opt",
                    "x86_64-slc5-gcc43-opt",
                    "slc5_amd64_gcc43",
                    "Linux_x86_64_glibc-2.12",
                    "Linux_x86_64_glibc-2.5",
                ],
                "UserPriority": 1,
                "Sites": [
                    "LCG.LAPP.fr",
                    "LCG.UKI-SOUTHGRID-RALPP.uk",
                ],
                "BannedSites": "LCG.KEK.jp",
                "JobTypes": "MCReconstruction_Overlay",
            },
            "Arguments": "jobDescription.xml -o LogLevel=verbose",
            "SoftwarePackages": [
                "overlayinput.1",
                "marlin.v0111Prod",
            ],
            "DebugLFNs": "",
            "Status": "Created",
            "InputDataModule": "DIRAC.WorkloadManagementSystem.Client.InputDataResolution",
            "BannedSites": "LCG.KEK.jp",
            "LogLevel": "verbose",
            "InputSandbox": [
                "jobDescription.xml",
                "SB:ProductionSandboxSE2|/SandBox/i/ilc_prod/5d3/92f/5d392f5266a796018ab6774ef84cbd31.tar.bz2",
            ],
            "OwnerName": "sailer",
            "StdOutput": "std.out",
            "JobType": "MCReconstruction_Overlay",
            "GridEnv": "/cvmfs/grid.cern.ch/emi-ui-3.7.3-1_sl6v2/etc/profile.d/setup-emi3-ui-example",
            "TransformationID": 6326,
            "DIRACSetup": "ILC-Production",
            "StdError": "std.err",
            "IS_PROD": "True",
            "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
            "JobGroup": 0o0006326,
            "OutputSandbox": [
                "std.err",
                "std.out",
            ],
            "JobID": 15756436,
            "VirtualOrganization": "ilc",
            "ProductionOutputData": [
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/000/yyveyx_o_rec_6326_15.slcio",
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/DST/00006326/000/yyveyx_o_dst_6326_15.slcio",
            ],
            "Site": "ANY",
            "OwnerGroup": "ilc_prod",
            "Owner": "sailer",
            "LogFilePath": "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/LOG/000",
            "InputData": "/ilc/prod/clic/500gev/yyveyx_o/ILD/SIM/00006325/000/yyveyx_o_sim_6325_17.slcio",
        }

        self.jdlBrokenContent = {
            "LogTargetPath": "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/LOG/00006326_015.tar",
            "Executable": "dirac-jobexec",
            "TaskID": "muahahaha",
            "SoftwareDistModule": "ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation",
            "JobName": "00006326_00000015",
            "Priority": 1,
            "Platform": "x86_64-slc5-gcc43-opt",
            "JobRequirements": {
                "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
                "VirtualOrganization": "ilc",
                "Setup": "ILC-Production",
                "CPUTime": 300000,
                "OwnerGroup": "ilc_prod",
                "Platforms": [
                    "x86_64-slc6-gcc44-opt",
                    "x86_64-slc5-gcc43-opt",
                    "slc5_amd64_gcc43",
                    "Linux_x86_64_glibc-2.12",
                    "Linux_x86_64_glibc-2.5",
                ],
                "UserPriority": 1,
                "Sites": [
                    "LCG.LAPP.fr",
                    "LCG.UKI-SOUTHGRID-RALPP.uk",
                ],
                "BannedSites": "LCG.KEK.jp",
                "JobTypes": "MCReconstruction_Overlay",
            },
            "Arguments": "jobDescription.xml -o LogLevel=verbose",
            "SoftwarePackages": [
                "overlayinput.1",
                "marlin.v0111Prod",
            ],
            "DebugLFNs": "",
            "Status": "Created",
            "InputDataModule": "DIRAC.WorkloadManagementSystem.Client.InputDataResolution",
            "BannedSites": "LCG.KEK.jp",
            "LogLevel": "verbose",
            "InputSandbox": [
                "jobDescription.xml",
                "SB:ProductionSandboxSE2|/SandBox/i/ilc_prod/5d3/92f/5d392f5266a796018ab6774ef84cbd31.tar.bz2",
            ],
            "OwnerName": "sailer",
            "StdOutput": "std.out",
            "JobType": "MCReconstruction_Overlay",
            "GridEnv": "/cvmfs/grid.cern.ch/emi-ui-3.7.3-1_sl6v2/etc/profile.d/setup-emi3-ui-example",
            "TransformationID": 6326,
            "DIRACSetup": "ILC-Production",
            "StdError": "std.err",
            "IS_PROD": "True",
            "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
            "JobGroup": 0o0006326,
            "OutputSandbox": [
                "std.err",
                "std.out",
            ],
            "JobID": 15756436,
            "VirtualOrganization": "ilc",
            "ProductionOutputData": [
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/000/yyveyx_o_rec_6326_15.slcio",
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/DST/00006326/000/yyveyx_o_dst_6326_15.slcio",
            ],
            "Site": "ANY",
            "OwnerGroup": "ilc_prod",
            "Owner": "sailer",
            "LogFilePath": "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/LOG/000",
            "InputData": "/ilc/prod/clic/500gev/yyveyx_o/ILD/SIM/00006325/000/yyveyx_o_sim_6325_17.slcio",
        }

        # jdl with single outputdata,
        self.jdl1 = {
            "LogTargetPath": "/ilc/prod/clic/3tev/e1e1_o/SID/SIM/00006301/LOG/00006301_10256.tar",
            "Executable": "dirac-jobexec",
            "TaskID": 10256,
            "SoftwareDistModule": "ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation",
            "JobName": "00006301_00010256",
            "Priority": 1,
            "Platform": "x86_64-slc5-gcc43-opt",
            "JobRequirements": {
                "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
                "VirtualOrganization": "ilc",
                "Setup": "ILC-Production",
                "CPUTime": 300000,
                "OwnerGroup": "ilc_prod",
                "Platforms": [
                    "x86_64-slc6-gcc44-opt",
                    "x86_64-slc5-gcc43-opt",
                    "slc5_amd64_gcc43",
                    "Linux_x86_64_glibc-2.12",
                    "Linux_x86_64_glibc-2.5",
                ],
                "UserPriority": 1,
                "Sites": [
                    "LCG.LAPP.fr",
                    "LCG.UKI-SOUTHGRID-RALPP.uk",
                ],
                "BannedSites": [
                    "OSG.MIT.us",
                    "OSG.SPRACE.br",
                ],
                "JobTypes": "MCSimulation",
            },
            "Arguments": "jobDescription.xml -o LogLevel=verbose",
            "SoftwarePackages": "slic.v2r9p8",
            "DebugLFNs": "",
            "Status": "Created",
            "InputDataModule": "DIRAC.WorkloadManagementSystem.Client.InputDataResolution",
            "BannedSites": [
                "OSG.MIT.us",
                "OSG.SPRACE.br",
            ],
            "LogLevel": "verbose",
            "InputSandbox": [
                "jobDescription.xml",
                "SB:ProductionSandboxSE2|/SandBox/i/ilc_prod/042/d64/042d64cb0fe73720cbd114a73506c582.tar.bz2",
            ],
            "OwnerName": "sailer",
            "StdOutput": "std.out",
            "JobType": "MCSimulation",
            "GridEnv": "/cvmfs/grid.cern.ch/emi-ui-3.7.3-1_sl6v2/etc/profile.d/setup-emi3-ui-example",
            "TransformationID": 6301,
            "DIRACSetup": "ILC-Production",
            "StdError": "std.err",
            "IS_PROD": "True",
            "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
            "JobGroup": "00006301",
            "OutputSandbox": [
                "std.err",
                "std.out",
            ],
            "JobID": 15756456,
            "VirtualOrganization": "ilc",
            "ProductionOutputData": "/ilc/prod/clic/3tev/e1e1_o/SID/SIM/00006301/010/e1e1_o_sim_6301_10256.slcio",
            "Site": "ANY",
            "OwnerGroup": "ilc_prod",
            "Owner": "sailer",
            "LogFilePath": "/ilc/prod/clic/3tev/e1e1_o/SID/SIM/00006301/LOG/010",
            "InputData": "/ilc/prod/clic/3tev/e1e1_o/gen/00006300/004/e1e1_o_gen_6300_4077.stdhep",
        }

        self.jdlNoInput = {
            "LogTargetPath": "/ilc/prod/clic/1.4tev/ea_qqqqnu/gen/00006498/LOG/00006498_1307.tar",
            "Executable": "dirac-jobexec",
            "TaskID": 1307,
            "SoftwareDistModule": "ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation",
            "JobName": "00006498_00001307",
            "Priority": 1,
            "Platform": "x86_64-slc5-gcc43-opt",
            "JobRequirements": {
                "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
                "VirtualOrganization": "ilc",
                "Setup": "ILC-Production",
                "CPUTime": 300000,
                "OwnerGroup": "ilc_prod",
                "Platforms": [
                    "x86_64-slc6-gcc44-opt",
                    "x86_64-slc5-gcc43-opt",
                    "slc5_amd64_gcc43",
                    "Linux_x86_64_glibc-2.12",
                    "Linux_x86_64_glibc-2.5",
                ],
                "UserPriority": 1,
                "BannedSites": "LCG.KEK.jp",
                "JobTypes": "MCGeneration",
            },
            "Arguments": "jobDescription.xml -o LogLevel=verbose",
            "SoftwarePackages": "whizard.SM_V57",
            "DebugLFNs": "",
            "Status": "Created",
            "InputDataModule": "DIRAC.WorkloadManagementSystem.Client.InputDataResolution",
            "BannedSites": "LCG.KEK.jp",
            "LogLevel": "verbose",
            "InputSandbox": [
                "jobDescription.xml",
                "SB:ProductionSandboxSE2|/SandBox/i/ilc_prod/b2a/d98/b2ad98c3e240361a4253c4bb277be478.tar.bz2",
            ],
            "OwnerName": "sailer",
            "StdOutput": "std.out",
            "JobType": "MCGeneration",
            "GridEnv": "/cvmfs/grid.cern.ch/emi-ui-3.7.3-1_sl6v2/etc/profile.d/setup-emi3-ui-example",
            "TransformationID": 6498,
            "DIRACSetup": "ILC-Production",
            "StdError": "std.err",
            "IS_PROD": "True",
            "OwnerDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sailer/CN=683529/CN=Andre Sailer",
            "JobGroup": "00006498",
            "OutputSandbox": [
                "std.err",
                "std.out",
            ],
            "JobID": 15762268,
            "VirtualOrganization": "ilc",
            "ProductionOutputData": "/ilc/prod/clic/1.4tev/ea_qqqqnu/gen/00006498/001/ea_qqqqnu_gen_6498_1307.stdhep",
            "Site": "ANY",
            "OwnerGroup": "ilc_prod",
            "Owner": "sailer",
            "LogFilePath": "/ilc/prod/clic/1.4tev/ea_qqqqnu/gen/00006498/LOG/001",
            "InputData": "",
        }

    def tearDown(self):
        pass

    def test_Init(self):
        """Transformation.Utilities.JobInfo init ...................................................."""
        assert self.jbi.outputFiles == []
        self.assertFalse(self.jbi.pendingRequest)

    def test_allFilesExist(self):
        """Transformation.Utilities.JobInfo.allFilesExist............................................"""
        self.jbi.outputFileStatus = ["Exists", "Exists"]
        self.assertTrue(self.jbi.allFilesExist())
        self.jbi.outputFileStatus = ["Exists", "Missing"]
        self.assertFalse(self.jbi.allFilesExist())
        self.jbi.outputFileStatus = ["Missing", "Exists"]
        self.assertFalse(self.jbi.allFilesExist())
        self.jbi.outputFileStatus = ["Missing", "Missing"]
        self.assertFalse(self.jbi.allFilesExist())
        self.jbi.outputFileStatus = []
        self.assertFalse(self.jbi.allFilesExist())

    def test_allFilesMissing(self):
        """Transformation.Utilities.JobInfo.allFilesMissing.........................................."""
        self.jbi.outputFileStatus = ["Exists", "Exists"]
        self.assertFalse(self.jbi.allFilesMissing())
        self.jbi.outputFileStatus = ["Exists", "Missing"]
        self.assertFalse(self.jbi.allFilesMissing())
        self.jbi.outputFileStatus = ["Missing", "Exists"]
        self.assertFalse(self.jbi.allFilesMissing())
        self.jbi.outputFileStatus = ["Missing", "Missing"]
        self.assertTrue(self.jbi.allFilesMissing())
        self.jbi.outputFileStatus = []
        self.assertFalse(self.jbi.allFilesMissing())

    @parameterized.expand(
        [
            ("someFilesMissing", "outputFileStatus", ["Exists", "Exists"], False),
            ("someFilesMissing", "outputFileStatus", ["Exists", "Missing"], True),
            ("someFilesMissing", "outputFileStatus", ["Missing", "Exists"], True),
            ("someFilesMissing", "outputFileStatus", ["Missing", "Missing"], False),
            ("someFilesMissing", "outputFileStatus", [], False),
            ("allInputFilesExist", "inputFileStatus", ["Exists", "Exists"], True),
            ("allInputFilesExist", "inputFileStatus", ["Exists", "Missing"], False),
            ("allInputFilesExist", "inputFileStatus", ["Missing", "Missing"], False),
            ("allInputFilesExist", "inputFileStatus", [], False),
            ("allInputFilesMissing", "inputFileStatus", ["Exists", "Exists"], False),
            ("allInputFilesMissing", "inputFileStatus", ["Exists", "Missing"], False),
            ("allInputFilesMissing", "inputFileStatus", ["Missing", "Missing"], True),
            ("allInputFilesMissing", "inputFileStatus", [], False),
            ("someInputFilesMissing", "inputFileStatus", ["Exists", "Exists"], False),
            ("someInputFilesMissing", "inputFileStatus", ["Exists", "Missing"], True),
            ("someInputFilesMissing", "inputFileStatus", ["Missing", "Exists"], True),
            ("someInputFilesMissing", "inputFileStatus", ["Missing", "Missing"], False),
            ("someInputFilesMissing", "inputFileStatus", [], False),
            ("allFilesProcessed", "transFileStatus", ["Processed", "Processed"], True),
            ("allFilesProcessed", "transFileStatus", ["Processed", "Assigned"], False),
            ("allFilesProcessed", "transFileStatus", ["Assigned", "Assigned"], False),
            ("allFilesProcessed", "transFileStatus", ["Deleted", "Deleted"], False),
            ("allFilesProcessed", "transFileStatus", ["Unused", "Unused"], False),
            ("allFilesProcessed", "transFileStatus", [], False),
            ("allFilesAssigned", "transFileStatus", ["Processed", "Processed"], True),
            ("allFilesAssigned", "transFileStatus", ["Processed", "Assigned"], True),
            ("allFilesAssigned", "transFileStatus", ["Assigned", "Assigned"], True),
            ("allFilesAssigned", "transFileStatus", ["Assigned", "Unused"], False),
            ("allFilesAssigned", "transFileStatus", ["Deleted", "Deleted"], False),
            ("allFilesAssigned", "transFileStatus", ["Unused", "Unused"], False),
            ("allFilesAssigned", "transFileStatus", [], False),
            ("checkErrorCount", "errorCounts", [0, 9], False),
            ("checkErrorCount", "errorCounts", [0, 10], False),
            ("checkErrorCount", "errorCounts", [0, 11], True),
            ("checkErrorCount", "errorCounts", [0, 12], True),
            ("allTransFilesDeleted", "transFileStatus", ["Deleted", "Deleted"], True),
            ("allTransFilesDeleted", "transFileStatus", ["Deleted", "Assigned"], False),
            ("allTransFilesDeleted", "transFileStatus", ["Assigned", "Deleted"], False),
            ("allTransFilesDeleted", "transFileStatus", ["Assigned", "Assigned"], False),
        ]
    )
    def test_fileChecker(self, func, attr, value, expected):
        setattr(self.jbi, attr, value)
        gLogger.notice("%s, %s, %s, %s, %s" % (getattr(self.jbi, func)(), func, attr, value, expected))
        assert expected == getattr(self.jbi, func)()

    def test_getJDL(self):
        """Transformation.Utilities.JobInfo.getJDL..................................................."""

        self.diracAPI.getJobJDL.return_value = S_OK(self.jdl1)
        jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        self.assertIsInstance(jdlList, dict)

        self.diracAPI.getJobJDL.return_value = S_ERROR("no mon")
        with self.assertRaises(RuntimeError) as contextManagedException:
            jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        self.assertIn("Failed to get jobJDL", str(contextManagedException.exception))

    def test_getTaskInfo_1(self):
        # task is only one
        wit = ["MCReconstruction"]
        self.jbi.taskID = 1234
        self.jbi.inputFiles = ["lfn"]
        tasksDict = {1234: [dict(FileID=123456, LFN="lfn", Status="Assigned", ErrorCount=7)]}
        lfnTaskDict = {}
        self.jbi.getTaskInfo(tasksDict, lfnTaskDict, wit)
        self.assertEqual(self.jbi.transFileStatus, ["Assigned"])
        self.assertEqual(self.jbi.otherTasks, [])

    def test_getTaskInfo_2(self):
        # there are other tasks
        wit = ["MCReconstruction"]
        self.jbi.taskID = 1234
        self.jbi.inputFiles = ["lfn"]
        tasksDict = {12: [dict(FileID=123456, LFN="lfn", Status="Processed", ErrorCount=7)]}
        lfnTaskDict = {"lfn": 12}
        self.jbi.getTaskInfo(tasksDict, lfnTaskDict, wit)
        self.assertEqual(self.jbi.transFileStatus, ["Processed"])
        self.assertEqual(self.jbi.otherTasks, [12])

    def test_getTaskInfo_3(self):
        # raise
        wit = ["MCReconstruction"]
        self.jbi.taskID = 1234
        self.jbi.inputFiles = ["otherLFN"]
        tasksDict = {1234: [dict(FileID=123456, LFN="lfn", Status="Processed", ErrorCount=23)]}
        lfnTaskDict = {}
        with self.assertRaisesRegexp(TaskInfoException, "InputFiles do not agree"):
            self.jbi.getTaskInfo(tasksDict, lfnTaskDict, wit)

    # def test_getTaskInfo_4(self):
    #   # raise keyError
    #   wit = ['MCReconstruction']
    #   self.jbi.taskID = 1235
    #   self.jbi.inputFiles = []
    #   tasksDict = {1234: dict(FileID=123456, LFN="lfn", Status="Processed")}
    #   lfnTaskDict = {}
    #   with self.assertRaisesRegex(KeyError, ""):
    #     self.jbi.getTaskInfo(tasksDict, lfnTaskDict, wit)

    def test_getTaskInfo_5(self):
        # raise inputFile
        wit = ["MCReconstruction"]
        self.jbi.taskID = 1235
        self.jbi.inputFiles = []
        tasksDict = {1234: dict(FileID=123456, LFN="lfn", Status="Processed")}
        lfnTaskDict = {}
        with self.assertRaisesRegexp(TaskInfoException, "InputFiles is empty"):
            self.jbi.getTaskInfo(tasksDict, lfnTaskDict, wit)

    def test_getJobInformation(self):
        """Transformation.Utilities.JobInfo.getJobInformation........................................"""
        self.diracAPI.getJobJDL.return_value = S_OK(self.jdl1)
        self.jbi.getJobInformation(self.diracAPI, self.jobMon)
        self.assertEqual(
            self.jbi.outputFiles, ["/ilc/prod/clic/3tev/e1e1_o/SID/SIM/00006301/010/e1e1_o_sim_6301_10256.slcio"]
        )
        self.assertEqual(10256, self.jbi.taskID)
        self.assertEqual(
            self.jbi.inputFiles, ["/ilc/prod/clic/3tev/e1e1_o/gen/00006300/004/e1e1_o_gen_6300_4077.stdhep"]
        )

        # empty jdl
        self.setUp()
        self.diracAPI.getJobJDL.return_value = S_OK({})
        self.jbi.getJobInformation(self.diracAPI, self.jobMon)
        self.assertEqual(self.jbi.outputFiles, [])
        self.assertIsNone(self.jbi.taskID)
        self.assertEqual(self.jbi.inputFiles, [])

    def test_getOutputFiles(self):
        """Transformation.Utilities.JobInfo.getOutputFiles..........................................."""
        # singleLFN
        self.diracAPI.getJobJDL.return_value = S_OK(self.jdl1)
        jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        self.jbi._JobInfo__getOutputFiles(jdlList)
        self.assertEqual(
            self.jbi.outputFiles, ["/ilc/prod/clic/3tev/e1e1_o/SID/SIM/00006301/010/e1e1_o_sim_6301_10256.slcio"]
        )

        # two LFNs
        self.diracAPI.getJobJDL.return_value = S_OK(self.jdl2)
        jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        self.jbi._JobInfo__getOutputFiles(jdlList)
        self.assertEqual(
            self.jbi.outputFiles,
            [
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/REC/00006326/000/yyveyx_o_rec_6326_15.slcio",
                "/ilc/prod/clic/500gev/yyveyx_o/ILD/DST/00006326/000/yyveyx_o_dst_6326_15.slcio",
            ],
        )

    def test_getTaskID(self):
        """Transformation.Utilities.JobInfo.getTaskID................................................"""
        # singleLFN
        self.diracAPI.getJobJDL.return_value = S_OK(self.jdl1)
        jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        self.jbi._JobInfo__getTaskID(jdlList)
        self.assertEqual(10256, self.jbi.taskID)

        # broken jdl
        out = StringIO()
        sys.stdout = out
        self.diracAPI.getJobJDL.return_value = S_OK(self.jdlBrokenContent)
        jdlList = self.jbi._JobInfo__getJDL(self.diracAPI)
        with self.assertRaises(ValueError):
            self.jbi._JobInfo__getTaskID(jdlList)

    def test_getInputFile(self):
        """Test the extraction of the inputFile from the JDL parameters."""
        # singleLFN
        self.jbi._JobInfo__getInputFile({"InputData": "/single/lfn2"})
        self.assertEqual(self.jbi.inputFiles, ["/single/lfn2"])

        # list with singleLFN
        self.jbi._JobInfo__getInputFile({"InputData": ["/single/lfn1"]})
        self.assertEqual(self.jbi.inputFiles, ["/single/lfn1"])

        # list with two LFN
        self.jbi._JobInfo__getInputFile({"InputData": ["/lfn1", "/lfn2"]})
        self.assertEqual(self.jbi.inputFiles, ["/lfn1", "/lfn2"])

    def test_checkFileExistence(self):
        """Transformation.Utilities.JobInfo.checkFileExistance......................................."""
        # input and output files
        repStatus = {"inputFile1": True, "inputFile2": False, "outputFile1": False, "outputFile2": True}
        self.jbi.inputFiles = ["inputFile1", "inputFile2", "inputFile3"]
        self.jbi.outputFiles = ["outputFile1", "outputFile2", "unknownFile"]
        self.jbi.checkFileExistence(repStatus)
        self.assertTrue(self.jbi.inputFilesExist[0])
        self.assertFalse(self.jbi.inputFilesExist[1])
        self.assertFalse(self.jbi.inputFilesExist[2])
        self.assertEqual(self.jbi.inputFileStatus, ["Exists", "Missing", "Unknown"])
        self.assertEqual(self.jbi.outputFileStatus, ["Missing", "Exists", "Unknown"])

        # just output files
        self.setUp()
        repStatus = {"inputFile": True, "outputFile1": False, "outputFile2": True}
        self.jbi.inputFiles = []
        self.jbi.outputFiles = ["outputFile1", "outputFile2", "unknownFile"]
        self.jbi.checkFileExistence(repStatus)
        self.assertEqual(self.jbi.outputFileStatus, ["Missing", "Exists", "Unknown"])

    @parameterized.expand(
        [
            param(
                ["123: Failed MCReconstruction Transformation: 1234 -- 5678 ", "inputFile (True, Assigned, Errors 0"],
                [],
            ),
            param(
                ["123: Failed MCReconstruction Transformation: 1234 -- 5678  (Last task [7777])"], [], otherTasks=[7777]
            ),
            param([], ["MCReconstruction Transformation"], trID=0, taID=0),
            param([], ["(Last task"], otherTasks=[]),
            param(
                ["PENDING REQUEST IGNORE THIS JOB"],
                [],
                pendingRequest=True,
            ),
            param(
                ["No Pending Requests"],
                [],
                pendingRequest=False,
            ),
        ]
    )
    def test__str__(self, asserts, assertNots, trID=1234, taID=5678, otherTasks=False, pendingRequest=False):
        jbi = JobInfo(jobID=123, status="Failed", tID=trID, tType="MCReconstruction")
        jbi.pendingRequest = pendingRequest
        jbi.otherTasks = otherTasks
        gLogger.notice("otherTasks: ", jbi.otherTasks)
        jbi.taskID = taID
        jbi.inputFiles = ["inputFile"]
        jbi.inputFilesExist = [True]
        jbi.transFileStatus = ["Assigned"]
        jbi.outputFiles = ["outputFile"]
        jbi.errorCounts = [0]
        info = str(jbi)
        for assertStr in asserts:
            self.assertIn(assertStr, info)
        for assertStr in assertNots:
            self.assertNotIn(assertStr, info)

    def test_TaskInfoException(self):
        """Transformation.Utilities.JobInfo.TaskInfoException........................................"""
        tie = TaskInfoException("notTasked")
        self.assertIsInstance(tie, Exception)
        self.assertIn("notTasked", str(tie))


if __name__ == "__main__":
    SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(TestJI)
    TESTRESULT = unittest.TextTestRunner(verbosity=3).run(SUITE)



########################################################################
# File :    HttpStorageAccessHandler.py
# Author :  A.T.
########################################################################

"""  The HttpStorageAccessHandler is a http server request handler to provide a secure http
     access to the DIRAC StorageElement and StorageElementProxy. It is derived from the
     SimpleHTTPRequestHandler standard python handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCS__ = "$Id$"

import os
import shutil
import random

from six.moves import BaseHTTPServer

from DIRAC.Core.Utilities.DictCache import DictCache


class HttpStorageAccessHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    register = DictCache()
    basePath = ""

    def do_GET(self):
        """Serve a GET request."""

        # Strip off leading slash
        key = self.path[1:]
        if not self.register.exists(key):
            self.send_error(401, "Invalid key provided, access denied")
            return None

        cache_path = self.register.get(key)
        fileList = os.listdir(cache_path)
        if len(fileList) == 1:
            path = os.path.join(cache_path, fileList[0])
        else:
            # multiple files, make archive
            unique = str(random.getrandbits(24))
            fileString = " ".join(fileList)
            os.system(
                "tar -cf %s/dirac_data_%s.tar --remove-files -C %s %s" % (cache_path, unique, cache_path, fileString)
            )
            path = os.path.join(cache_path, "dirac_data_%s.tar" % unique)

        f = self.send_head(path)
        if f:
            shutil.copyfileobj(f, self.wfile)
            f.close()
            self.register.delete(key)

    def send_head(self, path):
        """Prepare headers for the file download"""
        # path = self.translate_path(self.path)
        f = None
        try:
            # Always read in binary mode. Opening files in text mode may cause
            # newline translations, making the actual size of the content
            # transmitted *less* than the content-length!
            f = open(path, "rb")
        except IOError:
            self.send_error(404, "File not found")
            return None
        self.send_response(200)
        self.send_header("Content-type", "application/octet-stream")
        fs = os.fstat(f.fileno())
        self.send_header("Content-Length", str(fs[6]))
        self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
        fname = os.path.basename(path)
        self.send_header("Last-Modified", self.date_time_string(fs.st_mtime))
        self.send_header("Content-Disposition", "filename=%s" % fname)
        self.end_headers()
        return f



""" Accounting Cache
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os.path
import time
import threading

from DIRAC import S_OK, S_ERROR, gLogger, rootPath, gConfig
from DIRAC.Core.Utilities.DictCache import DictCache


class DataCache(object):
    def __init__(self, dirName="accountingPlots"):
        self.graphsLocation = os.path.join(gConfig.getValue("/LocalSite/InstancePath", rootPath), "data", dirName)
        self.cachedGraphs = {}
        self.alive = True
        self.purgeThread = threading.Thread(target=self.purgeExpired)
        self.purgeThread.setDaemon(1)
        self.purgeThread.start()
        self.__dataCache = DictCache()
        self.__graphCache = DictCache(deleteFunction=self._deleteGraph)
        self.__dataLifeTime = 600
        self.__graphLifeTime = 3600

    def setGraphsLocation(self, graphsDir):
        self.graphsLocation = graphsDir
        for graphName in os.listdir(self.graphsLocation):
            if graphName.find(".png") > 0:
                graphLocation = "%s/%s" % (self.graphsLocation, graphName)
                gLogger.verbose("Purging %s" % graphLocation)
                os.unlink(graphLocation)

    def purgeExpired(self):
        while self.alive:
            time.sleep(600)
            self.__graphCache.purgeExpired()
            self.__dataCache.purgeExpired()

    def getReportData(self, reportRequest, reportHash, dataFunc):
        """
        Get report data from cache if exists, else generate it
        """
        reportData = self.__dataCache.get(reportHash)
        if not reportData:
            retVal = dataFunc(reportRequest)
            if not retVal["OK"]:
                return retVal
            reportData = retVal["Value"]
            self.__dataCache.add(reportHash, self.__dataLifeTime, reportData)
        return S_OK(reportData)

    def getReportPlot(self, reportRequest, reportHash, reportData, plotFunc):
        """
        Get report data from cache if exists, else generate it
        """
        plotDict = self.__graphCache.get(reportHash)
        if not plotDict:
            basePlotFileName = "%s/%s" % (self.graphsLocation, reportHash)
            retVal = plotFunc(reportRequest, reportData, basePlotFileName)
            if not retVal["OK"]:
                return retVal
            plotDict = retVal["Value"]
            if plotDict["plot"]:
                plotDict["plot"] = "%s.png" % reportHash
            if plotDict["thumbnail"]:
                plotDict["thumbnail"] = "%s.thb.png" % reportHash
            self.__graphCache.add(reportHash, self.__graphLifeTime, plotDict)
        return S_OK(plotDict)

    def getPlotData(self, plotFileName):
        filename = "%s/%s" % (self.graphsLocation, plotFileName)
        try:
            fd = open(filename, "rb")
            data = fd.read()
            fd.close()
        except Exception as e:
            return S_ERROR("Can't open file %s: %s" % (plotFileName, str(e)))
        return S_OK(data)

    def _deleteGraph(self, plotDict):
        try:
            for key in plotDict:
                value = plotDict[key]
                if value:
                    fPath = os.path.join(self.graphsLocation, str(value))
                    if os.path.isfile(fPath):
                        gLogger.info("Deleting plot from cache", value)
                        os.unlink(fPath)
                    else:
                        gLogger.info("Plot has already been deleted", value)
        except Exception:
            pass



#!/bin/env python
"""
Create and put 'PutAndRegister' request with a single local file

  warning: make sure the file you want to put is accessible from DIRAC production hosts,
           i.e. put file on network fs (AFS or NFS), otherwise operation will fail!!!
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("requestName:  a request name")
    Script.registerArgument("LFN:          logical file name")
    Script.registerArgument("localFile:    local file you want to put")
    Script.registerArgument("targetSE:     target SE")
    Script.parseCommandLine()

    import DIRAC
    from DIRAC import gLogger

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    requestName, LFN, PFN, targetSE = Script.getPositionalArgs(group=True)

    if not os.path.isabs(LFN):
        gLogger.error("LFN should be absolute path!!!")
        DIRAC.exit(-1)

    gLogger.info(
        "will create request '%s' with 'PutAndRegister' "
        "operation using %s pfn and %s target SE" % (requestName, PFN, targetSE)
    )

    from DIRAC.RequestManagementSystem.Client.Request import Request
    from DIRAC.RequestManagementSystem.Client.Operation import Operation
    from DIRAC.RequestManagementSystem.Client.File import File
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.Core.Utilities.Adler import fileAdler

    if not os.path.exists(PFN):
        gLogger.error("%s does not exist" % PFN)
        DIRAC.exit(-1)
    if not os.path.isfile(PFN):
        gLogger.error("%s is not a file" % PFN)
        DIRAC.exit(-1)

    PFN = os.path.abspath(PFN)
    size = os.path.getsize(PFN)
    adler32 = fileAdler(PFN)

    request = Request()
    request.RequestName = requestName

    putAndRegister = Operation()
    putAndRegister.Type = "PutAndRegister"
    putAndRegister.TargetSE = targetSE
    opFile = File()
    opFile.LFN = LFN
    opFile.PFN = PFN
    opFile.Size = size
    opFile.Checksum = adler32
    opFile.ChecksumType = "ADLER32"
    putAndRegister.addFile(opFile)
    request.addOperation(putAndRegister)
    reqClient = ReqClient()
    putRequest = reqClient.putRequest(request)
    if not putRequest["OK"]:
        gLogger.error("unable to put request '%s': %s" % (requestName, putRequest["Message"]))
        DIRAC.exit(-1)

    gLogger.always("Request '%s' has been put to ReqDB for execution." % requestName)
    gLogger.always("You can monitor its status using command: 'dirac-rms-request %s'" % requestName)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" This runs the RMS scenari as unit test for the DB. For that,
it replaces the normal MySQL connection with an inmemory SQLite db
"""

# pylint: disable=invalid-name,wrong-import-position

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from mock import patch
from pytest import fixture

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from DIRAC import gLogger

from DIRAC.RequestManagementSystem.DB import RequestDB

from DIRAC.RequestManagementSystem.DB.test.RMSTestScenari import (
    test_dirty,
    test_scheduled,
    test_stress,
    test_stressBulk,
)


@fixture(scope="function")
def reqDB(request):
    """This fixture instanciate a RequestDB with an in memory sqlite backend"""

    def mock_requestDB__init__(self):
        """This mock creates the RequestDB with an in memory sqlite backend"""
        self.log = gLogger.getSubLogger("RequestDB")
        # Initialize the connection info
        self.engine = create_engine("sqlite:///:memory:", echo=False, pool_recycle=3600)
        RequestDB.metadata.bind = self.engine
        self.DBSession = sessionmaker(bind=self.engine)

    with patch.object(RequestDB.RequestDB, "__init__", mock_requestDB__init__):
        db = RequestDB.RequestDB()
        db.createTables()

        yield db



"""
  Provide uniform interface to backend for local and remote clients.return

  There's a pretty big assumption here: that DB and Handler expose the same calls, with identical signatures.
  This is not exactly the case for WMS DBs and services.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


def getDBOrClient(DB, serverName):
    """Tries to instantiate the DB object
    and returns it if we manage to connect to the DB,
    otherwise returns a Client of the server
    """
    from DIRAC import gLogger
    from DIRAC.Core.DISET.RPCClient import RPCClient

    try:
        myDB = DB()
        if myDB._connected:
            return myDB
    except Exception:
        pass

    gLogger.info("Can not connect to DB will use %s" % serverName)
    return RPCClient(serverName)


def getPilotAgentsDB():
    serverName = "WorkloadManagement/PilotManager"
    PilotAgentsDB = None
    try:
        from DIRAC.WorkloadManagementSystem.DB.PilotAgentsDB import PilotAgentsDB
    except Exception:
        pass
    return getDBOrClient(PilotAgentsDB, serverName)


pilotAgentsDB = getPilotAgentsDB()


def getVirtualMachineDB():
    serverName = "WorkloadManagement/VirtualMachineManager"
    VirtualMachineDB = None
    try:
        from DIRAC.WorkloadManagementSystem.DB.VirtualMachineDB import VirtualMachineDB
    except Exception:
        pass
    return getDBOrClient(VirtualMachineDB, serverName)


virtualMachineDB = getVirtualMachineDB()



""" AccountingCacheCommand

  The AccountingCacheCommand class is a command module that collects command
  classes to store accounting results in the accounting cache.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# FIXME: NOT Usable ATM
# missing doNew, doCache, doMaster

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.AccountingSystem.Client.ReportsClient import ReportsClient
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getCESiteMapping
from DIRAC.ResourceStatusSystem.Command.Command import Command


class SuccessfullJobsBySiteSplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(SuccessfullJobsBySiteSplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns successfull jobs using the DIRAC accounting system for every site
        for the last self.args[0] hours

        :params:
          :attr:`sites`: list of sites (when not given, take every site)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        sites = None
        if "sites" in self.args:
            sites = self.args["sites"]
        if sites is None:
            sites = getSites()
            if not sites["OK"]:
                return sites
            sites = sites["Value"]

        if not sites:
            return S_ERROR("Sites is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        successfulJobs = self.rClient.getReport(
            "Job", "NumberOfJobs", fromD, toD, {"FinalStatus": ["Done"], "Site": sites}, "Site"
        )
        if not successfulJobs["OK"]:
            return successfulJobs
        successfulJobs = successfulJobs["Value"]

        if "data" not in successfulJobs:
            return S_ERROR("Missing data key")
        if "granularity" not in successfulJobs:
            return S_ERROR("Missing granularity key")

        singlePlots = {}

        successfulJobs["data"] = {site: strToIntDict(value) for site, value in successfulJobs["data"].items()}

        for site, value in successfulJobs["data"].items():
            if site in sites:
                plot = {}
                plot["data"] = {site: value}
                plot["granularity"] = successfulJobs["granularity"]
                singlePlots[site] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class FailedJobsBySiteSplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(FailedJobsBySiteSplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns failed jobs using the DIRAC accounting system for every site
        for the last self.args[0] hours

        :params:
          :attr:`sites`: list of sites (when not given, take every site)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        sites = None
        if "sites" in self.args:
            sites = self.args["sites"]
        if sites is None:
            # FIXME: pointing to the CSHelper instead
            #      sources = self.rsClient.getSite( meta = {'columns': 'SiteName'} )
            #      if not sources[ 'OK' ]:
            #        return sources
            #      sources = [ si[0] for si in sources[ 'Value' ] ]
            sites = getSites()
            if not sites["OK"]:
                return sites
            sites = sites["Value"]

        if not sites:
            return S_ERROR("Sites is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        failedJobs = self.rClient.getReport(
            "Job", "NumberOfJobs", fromD, toD, {"FinalStatus": ["Failed"], "Site": sites}, "Site"
        )
        if not failedJobs["OK"]:
            return failedJobs
        failedJobs = failedJobs["Value"]

        if "data" not in failedJobs:
            return S_ERROR("Missing data key")
        if "granularity" not in failedJobs:
            return S_ERROR("Missing granularity key")

        failedJobs["data"] = {site: strToIntDict(value) for site, value in failedJobs["data"].items()}

        singlePlots = {}

        for site, value in failedJobs["data"].items():
            if site in sites:
                plot = {}
                plot["data"] = {site: value}
                plot["granularity"] = failedJobs["granularity"]
                singlePlots[site] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class SuccessfullPilotsBySiteSplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(SuccessfullPilotsBySiteSplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns successfull pilots using the DIRAC accounting system for every site
        for the last self.args[0] hours

        :params:
          :attr:`sites`: list of sites (when not given, take every site)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        sites = None
        if "sites" in self.args:
            sites = self.args["sites"]
        if sites is None:
            # FIXME: pointing to the CSHelper instead
            #      sources = self.rsClient.getSite( meta = {'columns': 'SiteName'} )
            #      if not sources[ 'OK' ]:
            #        return sources
            #      sources = [ si[0] for si in sources[ 'Value' ] ]
            sites = getSites()
            if not sites["OK"]:
                return sites
            sites = sites["Value"]

        if not sites:
            return S_ERROR("Sites is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        succesfulPilots = self.rClient.getReport(
            "Pilot", "NumberOfPilots", fromD, toD, {"GridStatus": ["Done"], "Site": sites}, "Site"
        )
        if not succesfulPilots["OK"]:
            return succesfulPilots
        succesfulPilots = succesfulPilots["Value"]

        if "data" not in succesfulPilots:
            return S_ERROR("Missing data key")
        if "granularity" not in succesfulPilots:
            return S_ERROR("Missing granularity key")

        succesfulPilots["data"] = {site: strToIntDict(value) for site, value in succesfulPilots["data"].items()}

        singlePlots = {}

        for site, value in succesfulPilots["data"].items():
            if site in sites:
                plot = {}
                plot["data"] = {site: value}
                plot["granularity"] = succesfulPilots["granularity"]
                singlePlots[site] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class FailedPilotsBySiteSplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(FailedPilotsBySiteSplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns failed jobs using the DIRAC accounting system for every site
        for the last self.args[0] hours

        :params:
          :attr:`sites`: list of sites (when not given, take every site)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        sites = None
        if "sites" in self.args:
            sites = self.args["sites"]
        if sites is None:
            # FIXME: pointing to the CSHelper instead
            #      sources = self.rsClient.getSite( meta = {'columns': 'SiteName'} )
            #      if not sources[ 'OK' ]:
            #        return sources
            #      sources = [ si[0] for si in sources[ 'Value' ] ]
            sites = getSites()
            if not sites["OK"]:
                return sites
            sites = sites["Value"]

        if not sites:
            return S_ERROR("Sites is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        failedPilots = self.rClient.getReport(
            "Pilot", "NumberOfPilots", fromD, toD, {"GridStatus": ["Aborted"], "Site": sites}, "Site"
        )
        if not failedPilots["OK"]:
            return failedPilots
        failedPilots = failedPilots["Value"]

        if "data" not in failedPilots:
            return S_ERROR("Missing data key")
        if "granularity" not in failedPilots:
            return S_ERROR("Missing granularity key")

        failedPilots["data"] = {site: strToIntDict(value) for site, value in failedPilots["data"].items()}

        singlePlots = {}

        for site, value in failedPilots["data"].items():
            if site in sites:
                plot = {}
                plot["data"] = {site: value}
                plot["granularity"] = failedPilots["granularity"]
                singlePlots[site] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class SuccessfullPilotsByCESplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(SuccessfullPilotsByCESplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

    def doCommand(self):
        """
        Returns successfull pilots using the DIRAC accounting system for every CE
        for the last self.args[0] hours

        :params:
          :attr:`CEs`: list of CEs (when not given, take every CE)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        ces = None
        if "ces" in self.args:
            ces = self.args["ces"]
        if ces is None:
            res = getCESiteMapping()
            if not res["OK"]:
                return res
            ces = list(res["Value"])

        if not ces:
            return S_ERROR("CEs is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        successfulPilots = self.rClient.getReport(
            "Pilot", "NumberOfPilots", fromD, toD, {"GridStatus": ["Done"], "GridCE": ces}, "GridCE"
        )
        if not successfulPilots["OK"]:
            return successfulPilots
        successfulPilots = successfulPilots["Value"]

        if "data" not in successfulPilots:
            return S_ERROR("Missing data key")
        if "granularity" not in successfulPilots:
            return S_ERROR("Missing granularity key")

        successfulPilots["data"] = {site: strToIntDict(value) for site, value in successfulPilots["data"].items()}

        singlePlots = {}

        for ce, value in successfulPilots["data"].items():
            if ce in ces:
                plot = {}
                plot["data"] = {ce: value}
                plot["granularity"] = successfulPilots["granularity"]
                singlePlots[ce] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class FailedPilotsByCESplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(FailedPilotsByCESplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

        if "ReportGenerator" in self.apis:
            self.rgClient = self.apis["ReportGenerator"]
        else:
            self.rgClient = RPCClient("Accounting/ReportGenerator")

        self.rClient.rpcClient = self.rgClient

    def doCommand(self):
        """
        Returns failed pilots using the DIRAC accounting system for every CE
        for the last self.args[0] hours

        :params:
          :attr:`CEs`: list of CEs (when not given, take every CE)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        ces = None
        if "ces" in self.args:
            ces = self.args["ces"]
        if ces is None:
            res = getCESiteMapping()
            if not res["OK"]:
                return res
            ces = list(res["Value"])

        if not ces:
            return S_ERROR("CEs is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        failedPilots = self.rClient.getReport(
            "Pilot", "NumberOfPilots", fromD, toD, {"GridStatus": ["Aborted"], "GridCE": ces}, "GridCE"
        )
        if not failedPilots["OK"]:
            return failedPilots
        failedPilots = failedPilots["Value"]

        if "data" not in failedPilots:
            return S_ERROR("Missing data key")
        if "granularity" not in failedPilots:
            return S_ERROR("Missing granularity key")

        failedPilots["data"] = {site: strToIntDict(value) for site, value in failedPilots["data"].items()}

        singlePlots = {}

        for ce, value in failedPilots["data"].items():
            if ce in ces:
                plot = {}
                plot["data"] = {ce: value}
                plot["granularity"] = failedPilots["granularity"]
                singlePlots[ce] = plot

        return S_OK(singlePlots)


################################################################################
################################################################################


class RunningJobsBySiteSplittedCommand(Command):
    def __init__(self, args=None, clients=None):

        super(RunningJobsBySiteSplittedCommand, self).__init__(args, clients)

        if "ReportsClient" in self.apis:
            self.rClient = self.apis["ReportsClient"]
        else:
            self.rClient = ReportsClient()

        if "ReportGenerator" in self.apis:
            self.rgClient = self.apis["ReportGenerator"]
        else:
            self.rgClient = RPCClient("Accounting/ReportGenerator")

        self.rClient.rpcClient = self.rgClient

    def doCommand(self):
        """
        Returns running and runned jobs, querying the WMSHistory
        for the last self.args[0] hours

        :params:
          :attr:`sites`: list of sites (when not given, take every sites)

        :returns:

        """

        if "hours" not in self.args:
            return S_ERROR("Number of hours not specified")
        hours = self.args["hours"]

        sites = None
        if "sites" in self.args:
            sites = self.args["sites"]
        if sites is None:
            # FIXME: pointing to the CSHelper instead
            #      sources = self.rsClient.getSite( meta = {'columns': 'SiteName'} )
            #      if not sources[ 'OK' ]:
            #        return sources
            #      sources = [ si[0] for si in sources[ 'Value' ] ]
            sites = getSites()
            if not sites["OK"]:
                return sites
            sites = sites["Value"]

        if not sites:
            return S_ERROR("Sites is empty")

        fromD = datetime.utcnow() - timedelta(hours=hours)
        toD = datetime.utcnow()

        runJobs = self.rClient.getReport("WMSHistory", "NumberOfJobs", fromD, toD, {}, "Site")
        if not runJobs["OK"]:
            return runJobs
        runJobs = runJobs["Value"]

        if "data" not in runJobs:
            return S_ERROR("Missing data key")
        if "granularity" not in runJobs:
            return S_ERROR("Missing granularity key")

        runJobs["data"] = {site: strToIntDict(value) for site, value in runJobs["data"].items()}

        singlePlots = {}

        for site, value in runJobs["data"].items():
            if site in sites:
                plot = {}
                plot["data"] = {site: value}
                plot["granularity"] = runJobs["granularity"]
                singlePlots[site] = plot

        return S_OK(singlePlots)



#!/usr/bin/env python
"""
Command to launch the Transformation Shell
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=False)

    from DIRAC.TransformationSystem.Client.TransformationCLI import TransformationCLI

    cli = TransformationCLI()
    cli.cmdloop()


if __name__ == "__main__":
    main()



""" The LSF TimeLeft utility interrogates the LSF batch system for the
    current CPU and Wallclock consumed, as well as their limits.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import time

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import runCommand
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage


from DIRAC.Core.Utilities.Os import sourceEnv

__RCSID__ = "$Id$"


class LSFResourceUsage(ResourceUsage):
    """
    This is the LSF plugin of the TimeLeft Utility
    """

    def __init__(self):
        """Standard constructor"""
        super(LSFResourceUsage, self).__init__("LSF", "LSB_JOBID")

        self.queue = os.environ.get("LSB_QUEUE")
        self.bin = os.environ.get("LSF_BINDIR")
        self.host = os.environ.get("LSB_HOSTS")
        self.year = time.strftime("%Y", time.gmtime())
        self.log.verbose(
            "LSB_JOBID=%s, LSB_QUEUE=%s, LSF_BINDIR=%s, LSB_HOSTS=%s" % (self.jobID, self.queue, self.bin, self.host)
        )

        self.cpuLimit = None
        self.cpuRef = None
        self.normRef = None
        self.wallClockLimit = None
        self.hostNorm = None

        cmd = "%s/bqueues -l %s" % (self.bin, self.queue)
        result = runCommand(cmd)
        if not result["OK"]:
            return

        lines = str(result["Value"]).split("\n")
        self.log.debug(
            "From %s" % cmd, "\n".join([line if len(line) <= 128 else line[:128] + " [...]" for line in lines])
        )
        for i in range(len(lines)):
            if re.search(".*CPULIMIT.*", lines[i]):
                info = lines[i + 1].split()
                if len(info) >= 4:
                    self.cpuLimit = float(info[0]) * 60
                    self.cpuRef = info[3]
                elif len(info) == 2 and info[1] == "min":
                    self.cpuLimit = float(info[0]) * 60
                    self.cpuRef = None
                else:
                    self.log.warn('Problem parsing "%s" for CPU limit' % lines[i + 1])
                    self.cpuLimit = -1
            elif re.search(".*RUNLIMIT.*", lines[i]):
                info = lines[i + 1].split()
                if len(info) >= 1:
                    self.wallClockLimit = float(info[0]) * 60
                else:
                    self.log.warn('Problem parsing "%s" for wall clock limit' % lines[i + 1])
                    self.wallClockLimit = -1

        modelMaxNorm = 0
        if self.cpuRef:
            # Now try to get the CPU_FACTOR for this reference CPU,
            # it must be either a Model, a Host or the largest Model

            cmd = "%s/lshosts -w %s" % (self.bin, self.cpuRef)
            result = runCommand(cmd)
            if result["OK"]:
                # At CERN this command will return an error since there is no host defined
                # with the name of the reference Host.
                lines = str(result["Value"]).split("\n")
                l1 = lines[0].split()
                l2 = lines[1].split()
                if len(l1) > len(l2):
                    self.log.error("Failed lshost command", "%s:\n %s\n %s" % (cmd, lines[0], lines[0]))
                else:
                    for i in range(len(l1)):
                        if l1[i] == "cpuf":
                            try:
                                self.normRef = float(l2[i])
                                self.log.info(
                                    "Reference Normalization taken from Host", "%s: %s" % (self.cpuRef, self.normRef)
                                )
                            except ValueError as e:
                                self.log.exception("Exception parsing lshosts output", "", e)

            if not self.normRef:
                # Try if there is a model define with the name of cpuRef
                cmd = "%s/lsinfo -m" % (self.bin)
                result = runCommand(cmd)
                if result["OK"]:
                    lines = str(result["Value"]).split("\n")
                    for line in lines[1:]:
                        words = line.split()
                        if len(words) > 1:
                            try:
                                norm = float(words[1])
                                if norm > modelMaxNorm:
                                    modelMaxNorm = norm
                                if words[0].find(self.cpuRef) > -1:
                                    self.normRef = norm
                                    self.log.info(
                                        "Reference Normalization taken from Host Model",
                                        "%s: %s" % (self.cpuRef, self.normRef),
                                    )
                            except ValueError as e:
                                self.log.exception("Exception parsing lsfinfo output", "", e)

            if not self.normRef:
                # Now parse LSF configuration files
                if not os.path.isfile("./lsf.sh"):
                    os.symlink(os.path.join(os.environ["LSF_ENVDIR"], "lsf.conf"), "./lsf.sh")
                # As the variables are not exported, we must force it
                ret = sourceEnv(10, ["./lsf", "&& export LSF_CONFDIR"])
                if ret["OK"]:
                    lsfEnv = ret["outputEnv"]
                    shared = None
                    try:
                        egoShared = os.path.join(lsfEnv["LSF_CONFDIR"], "ego.shared")
                        lsfShared = os.path.join(lsfEnv["LSF_CONFDIR"], "lsf.shared")
                        if os.path.exists(egoShared):
                            shared = egoShared
                        elif os.path.exists(lsfShared):
                            shared = lsfShared
                    except KeyError as e:
                        self.log.exception("Exception getting LSF configuration", "", e)
                    if shared:
                        with open(shared) as f:
                            hostModelSection = False
                            for line in f.readlines():
                                if line.find("Begin HostModel") == 0:
                                    hostModelSection = True
                                    continue
                                if not hostModelSection:
                                    continue
                                if line.find("End HostModel") == 0:
                                    break
                                line = line.strip()
                                if line and line.split()[0] == self.cpuRef:
                                    try:
                                        self.normRef = float(line.split()[1])
                                        self.log.info(
                                            "Reference Normalization taken from Configuration File",
                                            "(%s) %s: %s" % (shared, self.cpuRef, self.normRef),
                                        )
                                    except ValueError as e:
                                        self.log.exception("Exception reading LSF configuration", "", e)
                    else:
                        self.log.warn("Could not find LSF configuration")
                else:
                    self.log.error("Cannot source the LSF environment", ret["Message"])
        if not self.normRef:
            # If nothing works take this as the unit
            self.normRef = 1.0
            # If nothing worked, take the maximum defined for a Model
            # if modelMaxNorm:
            #  self.normRef = modelMaxNorm
            #  self.log.info('Reference Normalization taken from Max Model:', self.normRef)

        # Now get the Normalization for the current Host
        if self.host:
            cmd = "%s/lshosts -w %s" % (self.bin, self.host)
            result = runCommand(cmd)
            if result["OK"]:
                lines = str(result["Value"]).split("\n")
                l1 = lines[0].split()
                l2 = lines[1].split()
                if len(l1) > len(l2):
                    self.log.error("Failed lshost command", "%s:\n %s\n %s" % (cmd, lines[0], lines[0]))
                else:
                    for i in range(len(l1)):
                        if l1[i] == "cpuf":
                            try:
                                self.hostNorm = float(l2[i])
                                self.log.info("Host Normalization", "%s: %s" % (self.host, self.hostNorm))
                            except ValueError as e:
                                self.log.exception("Exception parsing lshosts output", l1, e)
                            finally:
                                break

            if self.hostNorm and self.normRef:
                self.hostNorm /= self.normRef
                self.log.info("CPU power w.r.t. batch unit", self.hostNorm)

            if self.hostNorm:
                # Set the limits in real seconds
                self.cpuLimit /= self.hostNorm
                self.wallClockLimit /= self.hostNorm

    #############################################################################
    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries CPU, CPULimit,
        WallClock, WallClockLimit, and Unit for current slot.
        """
        if not self.bin:
            return S_ERROR("Could not determine bin directory for LSF")
        if not self.hostNorm:
            return S_ERROR("Could not determine host Norm factor")

        cpu = None
        wallClock = None

        cmd = "%s/bjobs -W %s" % (self.bin, self.jobID)
        result = runCommand(cmd)
        if not result["OK"]:
            return result
        lines = str(result["Value"]).split("\n")
        l1 = lines[0].split()
        l2 = lines[1].split()
        if len(l1) > len(l2):
            self.log.error("Failed bjobs command", "%s:\n %s\n %s" % (cmd, lines[0], lines[0]))
            return S_ERROR("Can not parse LSF output")

        sCPU = None
        sStart = None
        for i in range(len(l1)):
            if l1[i] == "CPU_USED":
                sCPU = l2[i]
                lCPU = sCPU.split(":")
                try:
                    cpu = float(lCPU[0]) * 3600 + float(lCPU[1]) * 60 + float(lCPU[2])
                except (ValueError, IndexError) as _e:
                    pass
            elif l1[i] == "START_TIME":
                sStart = l2[i]
                sStart = "%s %s" % (sStart, self.year)
                try:
                    timeTup = time.strptime(sStart, "%m/%d-%H:%M:%S %Y")
                    wallClock = time.mktime(time.localtime()) - time.mktime(timeTup)
                except ValueError:
                    pass

        if cpu is None or wallClock is None:
            return S_ERROR("Failed to parse LSF output")

        consumed = {
            "CPU": cpu,
            "CPULimit": self.cpuLimit,
            "WallClock": wallClock,
            "WallClockLimit": self.wallClockLimit,
        }
        self.log.debug(consumed)

        if None not in consumed.values():
            return S_OK(consumed)
        else:
            missed = [key for key, val in consumed.items() if val is None]
            msg = "Could not determine some parameters"
            self.log.info(
                msg, ": %s\nThis is the stdout from the batch system call\n%s" % (",".join(missed), result["Value"])
            )
            return S_ERROR(msg)
